# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=invalid-name
""" Configuration file for AiiDA Import/Export module """
from __future__ import division
from __future__ import absolute_import
from __future__ import print_function

from aiida.orm import Computer, Group, GroupTypeString, Node, User, Log, Comment

__all__ = ('EXPORT_VERSION',)

# Current export version
EXPORT_VERSION = '0.7'

IMPORTGROUP_TYPE = GroupTypeString.IMPORTGROUP_TYPE.value
DUPL_SUFFIX = ' (Imported #{})'

# The name of the subfolder in which the node files are stored
NODES_EXPORT_SUBFOLDER = 'nodes'

# Giving names to the various entities. Attributes and links are not AiiDA
# entities but we will refer to them as entities in the file (to simplify
# references to them).
NODE_ENTITY_NAME = 'Node'
GROUP_ENTITY_NAME = 'Group'
COMPUTER_ENTITY_NAME = 'Computer'
USER_ENTITY_NAME = 'User'
LOG_ENTITY_NAME = 'Log'
COMMENT_ENTITY_NAME = 'Comment'

# The signatures used to reference the entities in the import/export file
NODE_SIGNATURE = 'aiida.backends.djsite.db.models.DbNode'
GROUP_SIGNATURE = 'aiida.backends.djsite.db.models.DbGroup'
COMPUTER_SIGNATURE = 'aiida.backends.djsite.db.models.DbComputer'
USER_SIGNATURE = 'aiida.backends.djsite.db.models.DbUser'
LOG_SIGNATURE = 'aiida.backends.djsite.db.models.DbLog'
COMMENT_SIGNATURE = 'aiida.backends.djsite.db.models.DbComment'

# Mapping from entity names to signatures (used by the SQLA import/export)
entity_names_to_signatures = {
    NODE_ENTITY_NAME: NODE_SIGNATURE,
    GROUP_ENTITY_NAME: GROUP_SIGNATURE,
    COMPUTER_ENTITY_NAME: COMPUTER_SIGNATURE,
    USER_ENTITY_NAME: USER_SIGNATURE,
    LOG_ENTITY_NAME: LOG_SIGNATURE,
    COMMENT_ENTITY_NAME: COMMENT_SIGNATURE
}

# Mapping from signatures to entity names (used by the SQLA import/export)
signatures_to_entity_names = {
    NODE_SIGNATURE: NODE_ENTITY_NAME,
    GROUP_SIGNATURE: GROUP_ENTITY_NAME,
    COMPUTER_SIGNATURE: COMPUTER_ENTITY_NAME,
    USER_SIGNATURE: USER_ENTITY_NAME,
    LOG_SIGNATURE: LOG_ENTITY_NAME,
    COMMENT_SIGNATURE: COMMENT_ENTITY_NAME
}

# Mapping from entity names to AiiDA classes (used by the SQLA import/export)
entity_names_to_entities = {
    NODE_ENTITY_NAME: Node,
    GROUP_ENTITY_NAME: Group,
    COMPUTER_ENTITY_NAME: Computer,
    USER_ENTITY_NAME: User,
    LOG_ENTITY_NAME: Log,
    COMMENT_ENTITY_NAME: Comment
}

# Mapping of entity names to SQLA class paths
entity_names_to_sqla_schema = {
    NODE_ENTITY_NAME: 'aiida.backends.sqlalchemy.models.node.DbNode',
    GROUP_ENTITY_NAME: 'aiida.backends.sqlalchemy.models.group.DbGroup',
    COMPUTER_ENTITY_NAME: 'aiida.backends.sqlalchemy.models.computer.DbComputer',
    USER_ENTITY_NAME: 'aiida.backends.sqlalchemy.models.user.DbUser',
    LOG_ENTITY_NAME: 'aiida.backends.sqlalchemy.models.log.DbLog',
    COMMENT_ENTITY_NAME: 'aiida.backends.sqlalchemy.models.comment.DbComment'
}

# Mapping of the export file fields (that coincide with the Django fields) to
# model fields that can be used for the query of the database in both backends.
# These are the names of the fields of the models that belong to the
# corresponding entities.

file_fields_to_model_fields = {
    NODE_ENTITY_NAME: {
        'dbcomputer': 'dbcomputer_id',
        'user': 'user_id'
    },
    GROUP_ENTITY_NAME: {
        'user': 'user_id'
    },
    COMPUTER_ENTITY_NAME: {},
    LOG_ENTITY_NAME: {
        'dbnode': 'dbnode_id'
    },
    COMMENT_ENTITY_NAME: {
        'dbnode': 'dbnode_id',
        'user': 'user_id'
    }
}
# As above but the opposite procedure
model_fields_to_file_fields = {
    NODE_ENTITY_NAME: {
        'dbcomputer_id': 'dbcomputer',
        'user_id': 'user'
    },
    GROUP_ENTITY_NAME: {
        'user_id': 'user'
    },
    COMPUTER_ENTITY_NAME: {},
    USER_ENTITY_NAME: {},
    LOG_ENTITY_NAME: {
        'dbnode_id': 'dbnode'
    },
    COMMENT_ENTITY_NAME: {
        'dbnode_id': 'dbnode',
        'user_id': 'user'
    }
}


def get_all_fields_info():
    """
    This method returns a description of the field names that should be used
    to describe the entity properties.
    Apart from of the listing of the fields per properties, it also shown
    the dependencies among different entities (and on which fields). It is
    also shown/return the unique identifiers used per entity.

    """
    unique_identifiers = {
        USER_ENTITY_NAME: 'email',
        COMPUTER_ENTITY_NAME: 'uuid',
        NODE_ENTITY_NAME: 'uuid',
        GROUP_ENTITY_NAME: 'uuid',
        LOG_ENTITY_NAME: 'uuid',
        COMMENT_ENTITY_NAME: 'uuid'
    }

    all_fields_info = dict()
    all_fields_info[USER_ENTITY_NAME] = {'last_name': {}, 'first_name': {}, 'institution': {}, 'email': {}}
    all_fields_info[COMPUTER_ENTITY_NAME] = {
        'transport_type': {},
        'hostname': {},
        'description': {},
        'scheduler_type': {},
        'metadata': {},
        'uuid': {},
        'name': {}
    }
    all_fields_info[NODE_ENTITY_NAME] = {
        'ctime': {
            'convert_type': 'date'
        },
        'uuid': {},
        'mtime': {
            'convert_type': 'date'
        },
        'node_type': {},
        'label': {},
        'user': {
            'requires': USER_ENTITY_NAME,
            'related_name': 'dbnodes'
        },
        'dbcomputer': {
            'requires': COMPUTER_ENTITY_NAME,
            'related_name': 'dbnodes'
        },
        'description': {},
        'process_type': {}
    }
    all_fields_info[GROUP_ENTITY_NAME] = {
        'description': {},
        'user': {
            'related_name': 'dbgroups',
            'requires': USER_ENTITY_NAME
        },
        'time': {
            'convert_type': 'date'
        },
        'type_string': {},
        'uuid': {},
        'label': {}
    }
    all_fields_info[LOG_ENTITY_NAME] = {
        'uuid': {},
        'time': {
            'convert_type': 'date'
        },
        'loggername': {},
        'levelname': {},
        'message': {},
        'metadata': {},
        'dbnode': {
            'related_name': 'dblogs',
            'requires': NODE_ENTITY_NAME
        }
    }
    all_fields_info[COMMENT_ENTITY_NAME] = {
        'uuid': {},
        'ctime': {
            'convert_type': 'date'
        },
        'mtime': {
            'convert_type': 'date'
        },
        'content': {},
        'dbnode': {
            'related_name': 'dbcomments',
            'requires': NODE_ENTITY_NAME
        },
        'user': {
            'related_name': 'dbcomments',
            'requires': USER_ENTITY_NAME
        }
    }
    return all_fields_info, unique_identifiers



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=too-many-lines
"""Unittests for REST API."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import tempfile

from aiida import orm
from aiida.backends.testbase import AiidaTestCase
from aiida.common import json
from aiida.common.links import LinkType
from aiida.restapi.api import App, AiidaApi


class RESTApiTestCase(AiidaTestCase):
    """
    Setup of the tests for the AiiDA RESTful-api
    """
    _url_prefix = '/api/v3'
    _dummy_data = {}
    _PERPAGE_DEFAULT = 20
    _LIMIT_DEFAULT = 400

    @classmethod
    def setUpClass(cls, *args, **kwargs):  # pylint: disable=too-many-locals, too-many-statements
        """
        Basides the standard setup we need to add few more objects in the
        database to be able to explore different requests/filters/orderings etc.
        """
        # call parent setUpClass method
        super(RESTApiTestCase, cls).setUpClass()

        # connect the app and the api
        # Init the api by connecting it the the app (N.B. respect the following
        # order, api.__init__)
        kwargs = dict(PREFIX=cls._url_prefix, PERPAGE_DEFAULT=cls._PERPAGE_DEFAULT, LIMIT_DEFAULT=cls._LIMIT_DEFAULT)

        cls.app = App(__name__)
        cls.app.config['TESTING'] = True
        AiidaApi(cls.app, **kwargs)

        # create test inputs
        cell = ((2., 0., 0.), (0., 2., 0.), (0., 0., 2.))
        structure = orm.StructureData(cell=cell)
        structure.append_atom(position=(0., 0., 0.), symbols=['Ba'])
        structure.store()

        cif = orm.CifData(ase=structure.get_ase())
        cif.store()

        parameter1 = orm.Dict(dict={'a': 1, 'b': 2})
        parameter1.store()

        parameter2 = orm.Dict(dict={'c': 3, 'd': 4})
        parameter2.store()

        kpoint = orm.KpointsData()
        kpoint.set_kpoints_mesh([4, 4, 4])
        kpoint.store()

        resources = {'num_machines': 1, 'num_mpiprocs_per_machine': 1}

        calcfunc = orm.CalcFunctionNode(computer=cls.computer)
        calcfunc.store()

        calc = orm.CalcJobNode(computer=cls.computer)
        calc.set_option('resources', resources)
        calc.set_attribute('attr1', 'OK')
        calc.set_attribute('attr2', 'OK')

        calc.add_incoming(structure, link_type=LinkType.INPUT_CALC, link_label='link_structure')
        calc.add_incoming(parameter1, link_type=LinkType.INPUT_CALC, link_label='link_parameter')

        aiida_in = 'The input file\nof the CalcJob node'
        # Add the calcjob_inputs folder with the aiida.in file to the CalcJobNode repository
        with tempfile.NamedTemporaryFile(mode='w+') as handle:
            handle.write(aiida_in)
            handle.flush()
            handle.seek(0)
            calc.put_object_from_filelike(handle, key='calcjob_inputs/aiida.in', force=True)
        calc.store()

        aiida_out = 'The output file\nof the CalcJob node'
        retrieved_outputs = orm.FolderData()
        # Add the calcjob_outputs folder with the aiida.out file to the FolderData node
        with tempfile.NamedTemporaryFile(mode='w+') as handle:
            handle.write(aiida_out)
            handle.flush()
            handle.seek(0)
            retrieved_outputs.put_object_from_filelike(handle, key='calcjob_outputs/aiida.out', force=True)
        retrieved_outputs.store()
        retrieved_outputs.add_incoming(calc, link_type=LinkType.CREATE, link_label='retrieved')

        kpoint.add_incoming(calc, link_type=LinkType.CREATE, link_label='create')

        calc1 = orm.CalcJobNode(computer=cls.computer)
        calc1.set_option('resources', resources)
        calc1.store()

        dummy_computers = [{
            'name': 'test1',
            'hostname': 'test1.epfl.ch',
            'transport_type': 'ssh',
            'scheduler_type': 'pbspro',
        }, {
            'name': 'test2',
            'hostname': 'test2.epfl.ch',
            'transport_type': 'ssh',
            'scheduler_type': 'torque',
        }, {
            'name': 'test3',
            'hostname': 'test3.epfl.ch',
            'transport_type': 'local',
            'scheduler_type': 'slurm',
        }, {
            'name': 'test4',
            'hostname': 'test4.epfl.ch',
            'transport_type': 'ssh',
            'scheduler_type': 'slurm',
        }]

        for dummy_computer in dummy_computers:
            computer = orm.Computer(**dummy_computer)
            computer.store()

        # Prepare typical REST responses
        cls.process_dummy_data()

    def get_dummy_data(self):
        return self._dummy_data

    def get_url_prefix(self):
        return self._url_prefix

    @classmethod
    def process_dummy_data(cls):
        # pylint: disable=fixme
        """
        This functions prepare atomic chunks of typical responses from the
        RESTapi and puts them into class attributes

        """
        # TODO: Storing the different nodes as lists and accessing them
        # by their list index is very fragile and a pain to debug.
        # Please change this!
        computer_projections = ['id', 'uuid', 'name', 'hostname', 'transport_type', 'scheduler_type']
        computers = orm.QueryBuilder().append(orm.Computer, tag='comp', project=computer_projections).order_by({
            'comp': [{
                'name': {
                    'order': 'asc'
                }
            }]
        }).dict()

        # Cast UUID into a string (e.g. in sqlalchemy it comes as a UUID object)
        computers = [_['comp'] for _ in computers]
        for comp in computers:
            if comp['uuid'] is not None:
                comp['uuid'] = str(comp['uuid'])
        cls._dummy_data['computers'] = computers

        calculation_projections = ['id', 'uuid', 'user_id', 'node_type']
        calculations = orm.QueryBuilder().append(orm.CalculationNode, tag='calc',
                                                 project=calculation_projections).order_by({
                                                     'calc': [{
                                                         'id': {
                                                             'order': 'desc'
                                                         }
                                                     }]
                                                 }).dict()

        calculations = [_['calc'] for _ in calculations]
        for calc in calculations:
            if calc['uuid'] is not None:
                calc['uuid'] = str(calc['uuid'])
        cls._dummy_data['calculations'] = calculations

        data_projections = ['id', 'uuid', 'user_id', 'node_type']
        data_types = {
            'cifdata': orm.CifData,
            'parameterdata': orm.Dict,
            'structuredata': orm.StructureData,
            'data': orm.Data,
        }
        for label, dataclass in data_types.items():
            data = orm.QueryBuilder().append(dataclass, tag='data', project=data_projections).order_by({
                'data': [{
                    'id': {
                        'order': 'desc'
                    }
                }]
            }).dict()
            data = [_['data'] for _ in data]

            for datum in data:
                if datum['uuid'] is not None:
                    datum['uuid'] = str(datum['uuid'])

            cls._dummy_data[label] = data

    def split_path(self, url):
        # pylint: disable=no-self-use
        """
        Split the url with "?" to get url path and it's parameters
        :param url: Web url
        :return: url path and url parameters
        """
        parts = url.split('?')
        path = ''
        query_string = ''
        if parts:
            path = parts[0]
        if len(parts) > 1:
            query_string = parts[1]

        return path, query_string

    def compare_extra_response_data(self, node_type, url, response, uuid=None):
        """
        In url response, we pass some extra information/data along with the node
        results. e.g. url method, node_type, path, pk, query_string, url,
        url_root,
        etc.

        :param node_type: url requested fot the type of the node
        :param url: web url
        :param response: url response
        :param uuid: url requested for the node pk
        """
        path, query_string = self.split_path(url)

        self.assertEqual(response['method'], 'GET')
        self.assertEqual(response['resource_type'], node_type)
        self.assertEqual(response['path'], path)
        self.assertEqual(response['id'], uuid)
        self.assertEqual(response['query_string'], query_string)
        self.assertEqual(response['url'], 'http://localhost' + url)
        self.assertEqual(response['url_root'], 'http://localhost/')

    # node details and list with limit, offset, page, perpage
    def process_test(
        self,
        entity_type,
        url,
        full_list=False,
        empty_list=False,
        expected_list_ids=None,
        expected_range=None,
        expected_errormsg=None,
        uuid=None,
        result_node_type=None,
        result_name=None
    ):
        # pylint: disable=too-many-arguments
        """
        Check whether response matches expected values.

        :param entity_type: url requested fot the type of the node
        :param url: web url
        :param full_list: if url is requested to get full list
        :param empty_list: if the response list is empty
        :param expected_list_ids: list of expected ids from data
        :param expected_range: [start, stop] range of expected ids from data
        :param expected_errormsg: expected error message in response
        :param uuid: url requested for the node pk
        :param result_node_type: node type in response data
        :param result_name: result name in response e.g. inputs, outputs
        """

        if expected_list_ids is None:
            expected_list_ids = []

        if expected_range is None:
            expected_range = []

        if result_node_type is None and result_name is None:
            result_node_type = entity_type
            result_name = entity_type

        url = self._url_prefix + url

        with self.app.test_client() as client:
            rv_response = client.get(url)
            response = json.loads(rv_response.data)

            if expected_errormsg:
                self.assertEqual(response['message'], expected_errormsg)
            else:
                if full_list:
                    expected_data = self._dummy_data[result_node_type]
                elif empty_list:
                    expected_data = []
                elif expected_list_ids:
                    expected_data = [self._dummy_data[result_node_type][i] for i in expected_list_ids]
                elif expected_range != []:
                    expected_data = self._dummy_data[result_node_type][expected_range[0]:expected_range[1]]
                else:
                    from aiida.common.exceptions import InputValidationError
                    raise InputValidationError('Pass the expected range of the dummydata')

                expected_node_uuids = [node['uuid'] for node in expected_data]
                result_node_uuids = [node['uuid'] for node in response['data'][result_name]]
                self.assertEqual(expected_node_uuids, result_node_uuids)

                self.compare_extra_response_data(entity_type, url, response, uuid)


class RESTApiTestSuite(RESTApiTestCase):
    # pylint: disable=too-many-public-methods
    """
    Define unittests for rest api
    """

    def test_computers_details(self):
        """
        Requests the details of single computer
        """
        node_uuid = self.get_dummy_data()['computers'][1]['uuid']
        RESTApiTestCase.process_test(
            self, 'computers', '/computers/' + str(node_uuid), expected_list_ids=[1], uuid=node_uuid
        )

    def test_computers_list(self):
        """
        Get the full list of computers from database
        """
        RESTApiTestCase.process_test(self, 'computers', '/computers?orderby=+id', full_list=True)

    def test_computers_list_limit_offset(self):
        """
        Get the list of computers from database using limit
        and offset parameter.
        It should return the no of rows specified in limit from
        database starting from the no. specified in offset
        """
        RESTApiTestCase.process_test(
            self, 'computers', '/computers?limit=2&offset=2&orderby=+id', expected_range=[2, 4]
        )

    def test_computers_list_limit_only(self):
        """
        Get the list of computers from database using limit
        parameter.
        It should return the no of rows specified in limit from
        database.
        """
        RESTApiTestCase.process_test(self, 'computers', '/computers?limit=2&orderby=+id', expected_range=[None, 2])

    def test_computers_list_offset_only(self):
        """
        Get the list of computers from database using offset
        parameter
        It should return all the rows from database starting from
        the no. specified in offset
        """
        RESTApiTestCase.process_test(self, 'computers', '/computers?offset=2&orderby=+id', expected_range=[2, None])

    def test_computers_list_limit_offset_perpage(self):
        """
        If we pass the limit, offset and perpage at same time, it
        would return the error message.
        """
        expected_error = 'perpage key is incompatible with limit and offset'
        RESTApiTestCase.process_test(
            self, 'computers', '/computers?offset=2&limit=1&perpage=2&orderby=+id', expected_errormsg=expected_error
        )

    def test_computers_list_page_limit_offset(self):
        """
        If we use the page, limit and offset at same time, it
        would return the error message.
        """
        expected_error = 'requesting a specific page is incompatible with ' \
                         'limit and offset'
        RESTApiTestCase.process_test(
            self, 'computers', '/computers/page/2?offset=2&limit=1&orderby=+id', expected_errormsg=expected_error
        )

    def test_complist_pagelimitoffset_perpage(self):
        """
        If we use the page, limit, offset and perpage at same time, it
        would return the error message.
        """
        expected_error = 'perpage key is incompatible with limit and offset'
        RESTApiTestCase.process_test(
            self,
            'computers',
            '/computers/page/2?offset=2&limit=1&perpage=2&orderby=+id',
            expected_errormsg=expected_error
        )

    def test_computers_list_page_default(self):
        """
        it returns the no. of rows defined as default perpage option
        from database.

        no.of pages = total no. of computers in database / perpage
        "/page" acts as "/page/1?perpage=default_value"

        """
        RESTApiTestCase.process_test(self, 'computers', '/computers/page?orderby=+id', full_list=True)

    def test_computers_list_page_perpage(self):
        """
        no.of pages = total no. of computers in database / perpage
        Using this formula it returns the no. of rows for requested page
        """
        RESTApiTestCase.process_test(
            self, 'computers', '/computers/page/1?perpage=2&orderby=+id', expected_range=[None, 2]
        )

    def test_computers_list_page_perpage_exceed(self):
        """
        no.of pages = total no. of computers in database / perpage

        If we request the page which exceeds the total no. of pages then
        it would return the error message.
        """
        expected_error = 'Non existent page requested. The page range is [1 : ' \
                         '3]'
        RESTApiTestCase.process_test(
            self, 'computers', '/computers/page/4?perpage=2&orderby=+id', expected_errormsg=expected_error
        )

    ############### list filters ########################
    def test_computers_filter_id1(self):
        """
        Add filter on the id of computer and get the filtered computer
        list (e.g. id=1)
        """
        node_pk = self.get_dummy_data()['computers'][1]['id']

        RESTApiTestCase.process_test(self, 'computers', '/computers?id=' + str(node_pk), expected_list_ids=[1])

    def test_computers_filter_id2(self):
        """
        Add filter on the id of computer and get the filtered computer
        list (e.g. id > 2)
        """
        node_pk = self.get_dummy_data()['computers'][1]['id']
        RESTApiTestCase.process_test(
            self, 'computers', '/computers?id>' + str(node_pk) + '&orderby=+id', expected_range=[2, None]
        )

    def test_computers_filter_pk(self):
        """
        Add filter on the id of computer and get the filtered computer
        list (e.g. id=1)
        """
        node_pk = self.get_dummy_data()['computers'][1]['id']
        RESTApiTestCase.process_test(self, 'computers', '/computers?pk=' + str(node_pk), expected_list_ids=[1])

    def test_computers_filter_name(self):
        """
        Add filter for the name of computer and get the filtered computer
        list
        """
        RESTApiTestCase.process_test(self, 'computers', '/computers?name="test1"', expected_list_ids=[1])

    def test_computers_filter_hostname(self):
        """
        Add filter for the hostname of computer and get the filtered computer
        list
        """
        RESTApiTestCase.process_test(self, 'computers', '/computers?hostname="test1.epfl.ch"', expected_list_ids=[1])

    def test_computers_filter_transport_type(self):
        """
        Add filter for the transport_type of computer and get the filtered
        computer
        list
        """
        RESTApiTestCase.process_test(
            self, 'computers', '/computers?transport_type="local"&name="test3"&orderby=+id', expected_list_ids=[3]
        )

    ############### list orderby ########################
    def test_computers_orderby_id_asc(self):
        """
        Returns the computers list ordered by "id" in ascending
        order
        """
        RESTApiTestCase.process_test(self, 'computers', '/computers?orderby=id', full_list=True)

    def test_computers_orderby_id_asc_sign(self):
        """
        Returns the computers list ordered by "+id" in ascending
        order
        """
        RESTApiTestCase.process_test(self, 'computers', '/computers?orderby=+id', full_list=True)

    def test_computers_orderby_id_desc(self):
        """
        Returns the computers list ordered by "id" in descending
        order
        """
        RESTApiTestCase.process_test(self, 'computers', '/computers?orderby=-id', expected_list_ids=[4, 3, 2, 1, 0])

    def test_computers_orderby_name_asc(self):
        """
        Returns the computers list ordered by "name" in ascending
        order
        """
        node_pk = self.get_dummy_data()['computers'][0]['id']
        RESTApiTestCase.process_test(
            self, 'computers', '/computers?pk>' + str(node_pk) + '&orderby=name', expected_list_ids=[1, 2, 3, 4]
        )

    def test_computers_orderby_name_asc_sign(self):
        """
        Returns the computers list ordered by "+name" in ascending
        order
        """
        node_pk = self.get_dummy_data()['computers'][0]['id']
        RESTApiTestCase.process_test(
            self, 'computers', '/computers?pk>' + str(node_pk) + '&orderby=+name', expected_list_ids=[1, 2, 3, 4]
        )

    def test_computers_orderby_name_desc(self):
        """
        Returns the computers list ordered by "name" in descending
        order
        """
        node_pk = self.get_dummy_data()['computers'][0]['id']
        RESTApiTestCase.process_test(
            self, 'computers', '/computers?pk>' + str(node_pk) + '&orderby=-name', expected_list_ids=[4, 3, 2, 1]
        )

    def test_computers_orderby_scheduler_type_asc(self):
        """
        Returns the computers list ordered by "scheduler_type" in ascending
        order
        """
        node_pk = self.get_dummy_data()['computers'][0]['id']
        RESTApiTestCase.process_test(
            self,
            'computers',
            '/computers?transport_type="ssh"&pk>' + str(node_pk) + '&orderby=scheduler_type',
            expected_list_ids=[1, 4, 2]
        )

    def test_comp_orderby_scheduler_ascsign(self):
        """
        Returns the computers list ordered by "+scheduler_type" in ascending
        order
        """
        node_pk = self.get_dummy_data()['computers'][0]['id']
        RESTApiTestCase.process_test(
            self,
            'computers',
            '/computers?transport_type="ssh"&pk>' + str(node_pk) + '&orderby=+scheduler_type',
            expected_list_ids=[1, 4, 2]
        )

    def test_computers_orderby_schedulertype_desc(self):
        """
        Returns the computers list ordered by "scheduler_type" in descending
        order
        """
        node_pk = self.get_dummy_data()['computers'][0]['id']
        RESTApiTestCase.process_test(
            self,
            'computers',
            '/computers?pk>' + str(node_pk) + '&transport_type="ssh"&orderby=-scheduler_type',
            expected_list_ids=[2, 4, 1]
        )

    ############### list orderby combinations #######################
    def test_computers_orderby_mixed1(self):
        """
        Returns the computers list first order by "transport_type" in
        ascending order and if it is having same transport_type, order it
        by "id"
        """
        node_pk = self.get_dummy_data()['computers'][0]['id']
        RESTApiTestCase.process_test(
            self,
            'computers',
            '/computers?pk>' + str(node_pk) + '&orderby=transport_type,id',
            expected_list_ids=[3, 1, 2, 4]
        )

    def test_computers_orderby_mixed2(self):
        """
        Returns the computers list first order by "scheduler_type" in
        descending order and if it is having same scheduler_type, order it
        by "name"
        """
        node_pk = self.get_dummy_data()['computers'][0]['id']
        RESTApiTestCase.process_test(
            self,
            'computers',
            '/computers?pk>' + str(node_pk) + '&orderby=-scheduler_type,name',
            expected_list_ids=[2, 3, 4, 1]
        )

    def test_computers_orderby_mixed3(self):
        """
        Returns the computers list first order by "scheduler_type" in
        ascending order and if it is having same scheduler_type, order it
        by "hostname" descending order

        Response::
        test4 slurm
        test3 slurm
        test2 torque
        test1 pbspro
        localhost pbspro
        ==========
        Expected::
        test1 pbspro
        localhost pbspro
        test4 slurm
        test3 slurm
        test2 torque
        test1 test4


        RESTApiTestCase.process_test(self, "computers",
                                  "/computers?orderby=+scheduler_type,
                                  -hostname",
                                  expected_list_ids=[1,0,4,3,2])
        """

    ############### list filter combinations #######################
    def test_computers_filter_mixed1(self):
        """
        Add filter for the hostname and id of computer and get the
        filtered computer list
        """
        node_pk = self.get_dummy_data()['computers'][0]['id']
        RESTApiTestCase.process_test(
            self, 'computers', '/computers?id>' + str(node_pk) + '&hostname="test1.epfl.ch"', expected_list_ids=[1]
        )

    def test_computers_filter_mixed2(self):
        """
        Add filter for the id, hostname and transport_type of the computer
        and get the filtered computer list
        """
        node_pk = self.get_dummy_data()['computers'][0]['id']
        RESTApiTestCase.process_test(
            self,
            'computers',
            '/computers?id>' + str(node_pk) + '&hostname="test3.epfl.ch"&transport_type="ssh"',
            empty_list=True
        )

    ############### list all parameter combinations #######################
    def test_computers_mixed1(self):
        """
        url parameters: id, limit and offset
        """
        node_pk = self.get_dummy_data()['computers'][0]['id']
        RESTApiTestCase.process_test(
            self, 'computers', '/computers?id>' + str(node_pk) + '&limit=2&offset=3', expected_list_ids=[4]
        )

    def test_computers_mixed2(self):
        """
        url parameters: id, page, perpage
        """
        node_pk = self.get_dummy_data()['computers'][0]['id']
        RESTApiTestCase.process_test(
            self,
            'computers',
            '/computers/page/2?id>' + str(node_pk) + '&perpage=2&orderby=+id',
            expected_list_ids=[3, 4]
        )

    def test_computers_mixed3(self):
        """
        url parameters: id, transport_type, orderby
        """
        node_pk = self.get_dummy_data()['computers'][0]['id']
        RESTApiTestCase.process_test(
            self,
            'computers',
            '/computers?id>=' + str(node_pk) + '&transport_type="ssh"&orderby=-id&limit=2',
            expected_list_ids=[4, 2]
        )

    ########## pass unknown url parameter ###########
    def test_computers_unknown_param(self):
        """
        url parameters: id, limit and offset

        from aiida.common.exceptions import InputValidationError
        RESTApiTestCase.node_exception(self, "/computers?aa=bb&id=2", InputValidationError)
        """

    ############### single calculation ########################
    def test_calculations_details(self):
        """
        Requests the details of single calculation
        """
        node_uuid = self.get_dummy_data()['calculations'][0]['uuid']
        RESTApiTestCase.process_test(
            self, 'calculations', '/calculations/' + str(node_uuid), expected_list_ids=[0], uuid=node_uuid
        )

    ############### full list with limit, offset, page, perpage #############
    def test_calculations_list(self):
        """
        Get the full list of calculations from database
        """
        RESTApiTestCase.process_test(self, 'calculations', '/calculations?orderby=-id', full_list=True)

    def test_calculations_list_limit_offset(self):
        """
        Get the list of calculations from database using limit
        and offset parameter.
        It should return the no of rows specified in limit from
        database starting from the no. specified in offset
        """
        RESTApiTestCase.process_test(
            self, 'calculations', '/calculations?limit=1&offset=1&orderby=+id', expected_list_ids=[1]
        )

    ############### calculation retrieved_inputs and retrieved_outputs  #############
    def test_calculation_retrieved_inputs(self):
        """
        Get the list of given calculation retrieved_inputs
        """
        node_uuid = self.get_dummy_data()['calculations'][1]['uuid']
        url = self.get_url_prefix() + '/calculations/' + str(node_uuid) + '/io/retrieved_inputs'
        with self.app.test_client() as client:
            response_value = client.get(url)
            response = json.loads(response_value.data)
            self.assertEqual(response['data']['retrieved_inputs'], ['calcjob_inputs/aiida.in'])

    def test_calculation_retrieved_outputs(self):
        """
        Get the list of given calculation retrieved_outputs
        """
        node_uuid = self.get_dummy_data()['calculations'][1]['uuid']
        url = self.get_url_prefix() + '/calculations/' + str(node_uuid) + '/io/retrieved_outputs'
        with self.app.test_client() as client:
            response_value = client.get(url)
            response = json.loads(response_value.data)
            self.assertEqual(response['data']['retrieved_outputs'], ['calcjob_outputs/aiida.out'])

    def test_calcfunction_retrieved_inputs(self):
        """
        Check that the given calcfunction does not have retrieved_inputs
        """
        node_uuid = self.get_dummy_data()['calculations'][2]['uuid']
        self.process_test(
            'calculations',
            '/calculations/' + str(node_uuid) + '/io/retrieved_inputs',
            uuid=node_uuid,
            result_name='retrieved_inputs',
            empty_list=True
        )

    def test_calcfunction_retrieved_outputs(self):
        """
        Check that the given calcfunction does not have retrieved_outputs
        """
        node_uuid = self.get_dummy_data()['calculations'][2]['uuid']
        self.process_test(
            'calculations',
            '/calculations/' + str(node_uuid) + '/io/retrieved_outputs',
            uuid=node_uuid,
            result_name='retrieved_outputs',
            empty_list=True
        )

    ############### calculation inputs  #############
    def test_calculation_inputs(self):
        """
        Get the list of give calculation inputs
        """
        node_uuid = self.get_dummy_data()['calculations'][1]['uuid']
        self.process_test(
            'calculations',
            '/calculations/' + str(node_uuid) + '/io/inputs?orderby=id',
            expected_list_ids=[5, 3],
            uuid=node_uuid,
            result_node_type='data',
            result_name='inputs'
        )

    def test_calculation_input_filters(self):
        """
        Get filtered inputs list for given calculations
        """
        node_uuid = self.get_dummy_data()['calculations'][1]['uuid']
        self.process_test(
            'calculations',
            '/calculations/' + str(node_uuid) + '/io/inputs?node_type="data.dict.Dict."',
            expected_list_ids=[3],
            uuid=node_uuid,
            result_node_type='data',
            result_name='inputs'
        )

    def test_calculation_iotree(self):
        """
        Get filtered inputs list for given calculations
        """
        node_uuid = self.get_dummy_data()['calculations'][1]['uuid']
        url = self.get_url_prefix() + '/calculations/' + str(node_uuid) + '/io/tree?in_limit=1&out_limit=1'
        with self.app.test_client() as client:
            response_value = client.get(url)
            response = json.loads(response_value.data)
            self.assertEqual(len(response['data']['nodes']), 3)
            self.assertEqual(len(response['data']['edges']), 2)
            expected_attr = ['nodelabel', 'nodetype', 'linklabel', 'linktype', 'nodeuuid', 'description']
            received_attr = response['data']['nodes'][1].keys()
            for attr in expected_attr:
                self.assertIn(attr, received_attr)
            RESTApiTestCase.compare_extra_response_data(self, 'calculations', url, response, uuid=node_uuid)

    ############### calculation attributes #############
    def test_calculation_attributes(self):
        """
        Get list of calculation attributes
        """
        attributes = {
            'attr1': 'OK',
            'attr2': 'OK',
            'resources': {
                'num_machines': 1,
                'num_mpiprocs_per_machine': 1
            },
        }
        node_uuid = self.get_dummy_data()['calculations'][1]['uuid']
        url = self.get_url_prefix() + '/calculations/' + str(node_uuid) + '/content/attributes'
        with self.app.test_client() as client:
            rv_obj = client.get(url)
            response = json.loads(rv_obj.data)
            self.assertNotIn('message', response)
            self.assertEqual(response['data']['attributes'], attributes)
            RESTApiTestCase.compare_extra_response_data(self, 'calculations', url, response, uuid=node_uuid)

    def test_calculation_attributes_nalist_filter(self):
        """
        Get list of calculation attributes with filter nalist
        """
        attributes = {
            'attr2': 'OK',
            'resources': {
                'num_machines': 1,
                'num_mpiprocs_per_machine': 1
            },
        }
        node_uuid = self.get_dummy_data()['calculations'][1]['uuid']
        url = self.get_url_prefix() + '/calculations/' + str(node_uuid) + '/content/attributes?nalist="attr1"'
        with self.app.test_client() as client:
            rv_obj = client.get(url)
            response = json.loads(rv_obj.data)
            self.assertNotIn('message', response)

            self.assertEqual(response['data']['attributes'], attributes)
            RESTApiTestCase.compare_extra_response_data(self, 'calculations', url, response, uuid=node_uuid)

    def test_calculation_attributes_alist_filter(self):
        """
        Get list of calculation attributes with filter alist
        """
        node_uuid = self.get_dummy_data()['calculations'][1]['uuid']
        url = self.get_url_prefix() + '/calculations/' + str(node_uuid) + '/content/attributes?alist="attr1"'
        with self.app.test_client() as client:
            rv_obj = client.get(url)
            response = json.loads(rv_obj.data)
            self.assertNotIn('message', response)
            self.assertEqual(response['data']['attributes'], {'attr1': 'OK'})
            RESTApiTestCase.compare_extra_response_data(self, 'calculations', url, response, uuid=node_uuid)

    ############### Structure visualization and download #############
    def test_structure_visualization(self):
        """
        Get the list of give calculation inputs
        """
        from aiida.backends.tests.test_dataclasses import simplify
        node_uuid = self.get_dummy_data()['structuredata'][0]['uuid']
        url = self.get_url_prefix() + '/structures/' + str(node_uuid) + '/content/visualization?visformat=cif'
        with self.app.test_client() as client:
            rv_obj = client.get(url)
            response = json.loads(rv_obj.data)
            self.assertNotIn('message', response)

            expected_visdata = """\n##########################################################################\n#               Crystallographic Information Format file \n#               Produced by PyCifRW module\n# \n#  This is a CIF file.  CIF has been adopted by the International\n#  Union of Crystallography as the standard for data archiving and \n#  transmission.\n#\n#  For information on this file format, follow the CIF links at\n#  http://www.iucr.org\n##########################################################################\n\ndata_0\nloop_\n  _atom_site_label\n  _atom_site_fract_x\n  _atom_site_fract_y\n  _atom_site_fract_z\n  _atom_site_type_symbol\n   Ba1  0.0  0.0  0.0  Ba\n \n_cell_angle_alpha                       90.0\n_cell_angle_beta                        90.0\n_cell_angle_gamma                       90.0\n_cell_length_a                          2.0\n_cell_length_b                          2.0\n_cell_length_c                          2.0\nloop_\n  _symmetry_equiv_pos_as_xyz\n   'x, y, z'\n \n_symmetry_int_tables_number             1\n_symmetry_space_group_name_H-M          'P 1'\n"""  # pylint: disable=line-too-long
            self.assertEqual(
                simplify(response['data']['visualization']['str_viz_info']['data']), simplify(expected_visdata)
            )
            self.assertEqual(response['data']['visualization']['str_viz_info']['format'], 'cif')
            self.assertEqual(
                response['data']['visualization']['dimensionality'], {
                    u'dim': 3,
                    u'value': 8.0,
                    u'label': u'volume'
                }
            )
            self.assertEqual(response['data']['visualization']['pbc'], [True, True, True])
            self.assertEqual(response['data']['visualization']['formula'], 'Ba')
            RESTApiTestCase.compare_extra_response_data(self, 'structures', url, response, uuid=node_uuid)

    def test_xsf_visualization(self):
        """
        Get the list of given calculation inputs
        """
        from aiida.backends.tests.test_dataclasses import simplify
        node_uuid = self.get_dummy_data()['structuredata'][0]['uuid']
        url = self.get_url_prefix() + '/structures/' + str(node_uuid) + '/content/visualization?visformat=xsf'
        with self.app.test_client() as client:
            rv_obj = client.get(url)
            response = json.loads(rv_obj.data)
            self.assertNotIn('message', response)

            expected_visdata = 'CRYSTAL\nPRIMVEC 1\n      2.0000000000       0.0000000000       0.0000000000\n      0.0000000000       2.0000000000       0.0000000000\n      0.0000000000       0.0000000000       2.0000000000\nPRIMCOORD 1\n1 1\n56       0.0000000000       0.0000000000       0.0000000000\n'  # pylint: disable=line-too-long
            self.assertEqual(
                simplify(response['data']['visualization']['str_viz_info']['data']), simplify(expected_visdata)
            )
            self.assertEqual(response['data']['visualization']['str_viz_info']['format'], 'xsf')
            self.assertEqual(
                response['data']['visualization']['dimensionality'], {
                    u'dim': 3,
                    u'value': 8.0,
                    u'label': u'volume'
                }
            )
            self.assertEqual(response['data']['visualization']['pbc'], [True, True, True])
            self.assertEqual(response['data']['visualization']['formula'], 'Ba')
            RESTApiTestCase.compare_extra_response_data(self, 'structures', url, response, uuid=node_uuid)

    def test_visualization(self):
        """
        Get the list of given calculation inputs
        """
        from aiida.backends.tests.test_dataclasses import simplify
        node_uuid = self.get_dummy_data()['structuredata'][0]['uuid']
        url = self.get_url_prefix() + '/structures/' + str(node_uuid) + '/content/visualization'
        with self.app.test_client() as client:
            rv_obj = client.get(url)
            response = json.loads(rv_obj.data)
            self.assertNotIn('message', response)

            expected_visdata = 'CRYSTAL\nPRIMVEC 1\n      2.0000000000       0.0000000000       0.0000000000\n      0.0000000000       2.0000000000       0.0000000000\n      0.0000000000       0.0000000000       2.0000000000\nPRIMCOORD 1\n1 1\n56       0.0000000000       0.0000000000       0.0000000000\n'  # pylint: disable=line-too-long
            self.assertEqual(
                simplify(response['data']['visualization']['str_viz_info']['data']), simplify(expected_visdata)
            )
            self.assertEqual(response['data']['visualization']['str_viz_info']['format'], 'xsf')
            self.assertEqual(
                response['data']['visualization']['dimensionality'], {
                    u'dim': 3,
                    u'value': 8.0,
                    u'label': u'volume'
                }
            )
            self.assertEqual(response['data']['visualization']['pbc'], [True, True, True])
            self.assertEqual(response['data']['visualization']['formula'], 'Ba')
            RESTApiTestCase.compare_extra_response_data(self, 'structures', url, response, uuid=node_uuid)

    def test_cif(self):
        """
        Test download of cif file
        """
        from aiida.orm import load_node

        node_uuid = self.get_dummy_data()['cifdata'][0]['uuid']
        url = self.get_url_prefix() + '/cifs/' + node_uuid + '/content/download'

        with self.app.test_client() as client:
            rv_obj = client.get(url)

        cif = load_node(node_uuid)._prepare_cif()[0]  # pylint: disable=protected-access
        self.assertEqual(rv_obj.data, cif)

    ############### schema #############
    def test_schema(self):
        """
        test schema
        """
        for nodetype in ['nodes', 'calculations', 'data', 'codes', 'computers', 'users', 'groups']:
            url = self.get_url_prefix() + '/' + nodetype + '/schema'
            with self.app.test_client() as client:
                rv_obj = client.get(url)
                response = json.loads(rv_obj.data)
                self.assertNotIn('message', response)

                expected_keys = ['display_name', 'help_text', 'is_display', 'is_foreign_key', 'type']

                # check fields
                for _, pinfo in response['data']['fields'].items():
                    available_keys = pinfo.keys()
                    for prop in expected_keys:
                        self.assertIn(prop, available_keys)

                # check order
                available_properties = response['data']['fields'].keys()
                for prop in response['data']['ordering']:
                    self.assertIn(prop, available_properties)

    def test_node_types(self):
        """
        Test the rest api call to get list of available node types
        """
        url = self.get_url_prefix() + '/nodes/types'
        with self.app.test_client() as client:
            rv_obj = client.get(url)
            response = json.loads(rv_obj.data)
            self.assertEqual(
                response['data']['process'],
                ['process.calculation.calcfunction.CalcFunctionNode.', 'process.calculation.calcjob.CalcJobNode.']
            )
            expected_data_types = [
                'data.array.kpoints.KpointsData.', 'data.cif.CifData.', 'data.dict.Dict.', 'data.folder.FolderData.',
                'data.structure.StructureData.'
            ]
            self.assertEqual(response['data']['data'], expected_data_types)
            RESTApiTestCase.compare_extra_response_data(self, 'nodes', url, response)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Extras tests for the export and import routines"""
# pylint: disable=attribute-defined-outside-init
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import with_statement

import os
import shutil
import tempfile

from aiida import orm
from aiida.backends.testbase import AiidaTestCase
from aiida.tools.importexport import import_data, export


class TestExtras(AiidaTestCase):
    """Test ex-/import cases related to Extras"""

    @classmethod
    def setUpClass(cls, *args, **kwargs):
        """Only run to prepare an export file"""
        super(TestExtras, cls).setUpClass()

        data = orm.Data()
        data.label = 'my_test_data_node'
        data.store()
        data.set_extra_many({'b': 2, 'c': 3})
        cls.tmp_folder = tempfile.mkdtemp()
        cls.export_file = os.path.join(cls.tmp_folder, 'export.aiida')
        export([data], outfile=cls.export_file, silent=True)

    @classmethod
    def tearDownClass(cls, *args, **kwargs):
        """Remove tmp_folder"""
        super(TestExtras, cls).tearDownClass()

        shutil.rmtree(cls.tmp_folder, ignore_errors=True)

    def setUp(self):
        """This function runs before every test execution"""
        self.clean_db()
        self.insert_data()

    def import_extras(self, mode_new='import'):
        """Import an aiida database"""
        import_data(self.export_file, silent=True, extras_mode_new=mode_new)

        builder = orm.QueryBuilder().append(orm.Data, filters={'label': 'my_test_data_node'})

        self.assertEqual(builder.count(), 1)
        self.imported_node = builder.all()[0][0]

    def modify_extras(self, mode_existing):
        """Import the same aiida database again"""
        self.imported_node.set_extra('a', 1)
        self.imported_node.set_extra('b', 1000)
        self.imported_node.delete_extra('c')

        import_data(self.export_file, silent=True, extras_mode_existing=mode_existing)

        # Query again the database
        builder = orm.QueryBuilder().append(orm.Data, filters={'label': 'my_test_data_node'})
        self.assertEqual(builder.count(), 1)
        return builder.all()[0][0]

    def tearDown(self):
        pass

    def test_import_of_extras(self):
        """Check if extras are properly imported"""
        self.import_extras()
        self.assertEqual(self.imported_node.get_extra('b'), 2)
        self.assertEqual(self.imported_node.get_extra('c'), 3)

    def test_absence_of_extras(self):
        """Check whether extras are not imported if the mode is set to 'none'"""
        self.import_extras(mode_new='none')
        with self.assertRaises(AttributeError):
            # the extra 'b' should not exist
            self.imported_node.get_extra('b')
        with self.assertRaises(AttributeError):
            # the extra 'c' should not exist
            self.imported_node.get_extra('c')

    def test_extras_import_mode_keep_existing(self):
        """Check if old extras are not modified in case of name collision"""
        self.import_extras()
        imported_node = self.modify_extras(mode_existing='kcl')

        # Check that extras are imported correctly
        self.assertEqual(imported_node.get_extra('a'), 1)
        self.assertEqual(imported_node.get_extra('b'), 1000)
        self.assertEqual(imported_node.get_extra('c'), 3)

    def test_extras_import_mode_update_existing(self):
        """Check if old extras are modified in case of name collision"""
        self.import_extras()
        imported_node = self.modify_extras(mode_existing='kcu')

        # Check that extras are imported correctly
        self.assertEqual(imported_node.get_extra('a'), 1)
        self.assertEqual(imported_node.get_extra('b'), 2)
        self.assertEqual(imported_node.get_extra('c'), 3)

    def test_extras_import_mode_mirror(self):
        """Check if old extras are fully overwritten by the imported ones"""
        self.import_extras()
        imported_node = self.modify_extras(mode_existing='ncu')

        # Check that extras are imported correctly
        with self.assertRaises(AttributeError):  # the extra
            # 'a' should not exist, as the extras were fully mirrored with respect to
            # the imported node
            imported_node.get_extra('a')
        self.assertEqual(imported_node.get_extra('b'), 2)
        self.assertEqual(imported_node.get_extra('c'), 3)

    def test_extras_import_mode_none(self):
        """Check if old extras are fully overwritten by the imported ones"""
        self.import_extras()
        imported_node = self.modify_extras(mode_existing='knl')

        # Check if extras are imported correctly
        self.assertEqual(imported_node.get_extra('b'), 1000)
        self.assertEqual(imported_node.get_extra('a'), 1)
        with self.assertRaises(AttributeError):  # the extra
            # 'c' should not exist, as the extras were keept untached
            imported_node.get_extra('c')

    def test_extras_import_mode_strange(self):
        """Check a mode that is probably does not make much sense but is still available"""
        self.import_extras()
        imported_node = self.modify_extras(mode_existing='kcd')

        # Check if extras are imported correctly
        self.assertEqual(imported_node.get_extra('a'), 1)
        self.assertEqual(imported_node.get_extra('c'), 3)
        with self.assertRaises(AttributeError):  # the extra
            # 'b' should not exist, as the collided extras are deleted
            imported_node.get_extra('b')

    def test_extras_import_mode_correct(self):
        """Test all possible import modes except 'ask' """
        self.import_extras()
        for mode1 in ['k', 'n']:  # keep or not keep old extras
            for mode2 in ['n', 'c']:  # create or not create new extras
                for mode3 in ['l', 'u', 'd']:  # leave old, update or delete collided extras
                    mode = mode1 + mode2 + mode3
                    import_data(self.export_file, silent=True, extras_mode_existing=mode)

    def test_extras_import_mode_wrong(self):
        """Check a mode that is wrong"""
        from aiida.tools.importexport.common.exceptions import ImportValidationError

        self.import_extras()
        with self.assertRaises(ImportValidationError):
            import_data(self.export_file, silent=True, extras_mode_existing='xnd')  # first letter is wrong
        with self.assertRaises(ImportValidationError):
            import_data(self.export_file, silent=True, extras_mode_existing='nxd')  # second letter is wrong
        with self.assertRaises(ImportValidationError):
            import_data(self.export_file, silent=True, extras_mode_existing='nnx')  # third letter is wrong
        with self.assertRaises(ImportValidationError):
            import_data(self.export_file, silent=True, extras_mode_existing='n')  # too short
        with self.assertRaises(ImportValidationError):
            import_data(self.export_file, silent=True, extras_mode_existing='nndnn')  # too long
        with self.assertRaises(ImportValidationError):
            import_data(self.export_file, silent=True, extras_mode_existing=5)  # wrong type



#!/usr/bin/env python
# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
""" Validate consistency of versions and dependencies.

Validates consistency of setup.json and

 * environment.yml
 * version in aiida/__init__.py
 * reentry dependency in pyproject.toml

"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import io
import os
import sys
import json
from collections import OrderedDict
import toml
import click

FILENAME_TOML = 'pyproject.toml'
FILENAME_SETUP_JSON = 'setup.json'
SCRIPT_PATH = os.path.split(os.path.realpath(__file__))[0]
ROOT_DIR = os.path.join(SCRIPT_PATH, os.pardir)
FILEPATH_SETUP_JSON = os.path.join(ROOT_DIR, FILENAME_SETUP_JSON)
FILEPATH_TOML = os.path.join(ROOT_DIR, FILENAME_TOML)


def get_setup_json():
    """Return the `setup.json` as a python dictionary """
    with open(FILEPATH_SETUP_JSON, 'r') as fil:
        setup_json = json.load(fil, object_pairs_hook=OrderedDict)

    return setup_json


def determine_block_positions(lines, block_start_marker, block_end_marker):
    """Determine the line indices of a block in a list of lines indicated by a given start and end marker.

    :param lines: list of strings
    :param block_start_marker: string marking the beginning of the block
    :param block_end_marker: string marking the end of the block
    :return: tuple of two integers representing the line indices that indicate the start and end of the block
    """
    block_start_index = -1
    block_end_index = -1

    for line_number, line in enumerate(lines):
        if block_start_marker in line:
            block_start_index = line_number + 1

        if block_end_marker in line:
            block_end_index = line_number
            break

    if block_start_index < 0 or block_end_index < 0:
        raise RuntimeError('failed to determine the starting or end point of the block')

    return block_start_index, block_end_index


def replace_line_block(lines, block, index_start, index_end):
    """Replace a block of lines between two line indices with a new set of lines.

    :param lines: list of lines representing the whole file
    :param block: list of lines representing the new block that should be inserted after old block is removed
    :param index_start: start of the block to be removed
    :param index_end: end of the block to be removed
    :return: list of lines with block of lines replaced
    """
    # Slice out the old block by removing the lines between the markers of the block
    lines = lines[:index_start] + lines[index_end:]

    # Now insert the new block starting at the beginning of the original block
    lines[index_start:index_start] = block

    return lines


def replace_block_in_file(filepath, block_start_marker, block_end_marker, block):
    """Replace a block of text between the given string markers with the provided new block of lines.

    :param filepath: absolute path of the file
    :param block_start_marker: string marking the beginning of the block
    :param block_end_marker: string marking the end of the block
    :param block: list of lines representing the new block that should be inserted after old block is removed
    """
    with io.open(filepath) as handle:
        lines = handle.readlines()

    try:
        index_start, index_end = determine_block_positions(lines, block_start_marker, block_end_marker)
    except RuntimeError as exception:
        raise RuntimeError('problem rewriting file `{}`:: {}'.format(filepath, exception))

    lines = replace_line_block(lines, block, index_start, index_end)

    with io.open(filepath, 'w') as handle:
        for line in lines:
            handle.write(line)


@click.group()
def cli():
    pass


@cli.command('verdi-autodocs')
def validate_verdi_documentation():
    """Auto-generate the documentation for `verdi` through `click`."""
    from click import Context
    from aiida.cmdline.commands.cmd_verdi import verdi

    # Set the `verdi data` command to isolated mode such that external plugin commands are not discovered
    ctx = Context(verdi)
    command = verdi.get_command(ctx, 'data')
    command.set_exclude_external_plugins(True)

    # Replacing the block with the overview of `verdi`
    filepath_verdi_overview = os.path.join(ROOT_DIR, 'docs', 'source', 'working_with_aiida', 'index.rst')
    overview_block_start_marker = '.. _verdi_overview:'
    overview_block_end_marker = '.. END_OF_VERDI_OVERVIEW_MARKER'

    # Generate the new block with the command index
    block = []
    for name, command in sorted(verdi.commands.items()):
        short_help = command.help.split('\n')[0]
        block.append(u'* :ref:`{name:}<verdi_{name:}>`:  {help:}\n'.format(name=name, help=short_help))

    # New block should start and end with an empty line after and before the literal block marker
    block.insert(0, u'\n')
    block.append(u'\n')

    replace_block_in_file(filepath_verdi_overview, overview_block_start_marker, overview_block_end_marker, block)

    # Replacing the block with the commands of `verdi`
    filepath_verdi_commands = os.path.join(ROOT_DIR, 'docs', 'source', 'verdi', 'verdi_user_guide.rst')
    commands_block_start_marker = '.. _verdi_commands:'
    commands_block_end_marker = '.. END_OF_VERDI_COMMANDS_MARKER'

    # Generate the new block with the command help strings
    header = u'Commands'
    message = u'Below is a list with all available subcommands.'
    block = [u'{}\n{}\n{}\n\n'.format(header, '=' * len(header), message)]

    for name, command in sorted(verdi.commands.items()):
        ctx = click.Context(command)

        header_label = u'.. _verdi_{name:}:'.format(name=name)
        header_string = u'``verdi {name:}``'.format(name=name)
        header_underline = u'-' * len(header_string)

        block.append(header_label + '\n\n')
        block.append(header_string + '\n')
        block.append(header_underline + '\n\n')
        block.append(u'::\n\n')  # Mark the beginning of a literal block
        for line in ctx.get_help().split('\n'):
            if line:
                block.append(u'    {}\n'.format(line))
            else:
                block.append(u'\n')
        block.append(u'\n\n')

    # New block should start and end with an empty line after and before the literal block marker
    block.insert(0, u'\n')
    block.append(u'\n')

    replace_block_in_file(filepath_verdi_commands, commands_block_start_marker, commands_block_end_marker, block)


@cli.command('version')
def validate_version():
    """Check that version numbers match.

    Check version number in setup.json and aiida-core/__init__.py and make sure
    they match.
    """
    # Get version from python package
    sys.path.insert(0, ROOT_DIR)
    import aiida  # pylint: disable=wrong-import-position
    version = aiida.__version__

    setup_content = get_setup_json()
    if version != setup_content['version']:
        click.echo('Version number mismatch detected:')
        click.echo("Version number in '{}': {}".format(FILENAME_SETUP_JSON, setup_content['version']))
        click.echo("Version number in '{}/__init__.py': {}".format('aiida', version))
        click.echo("Updating version in '{}' to: {}".format(FILENAME_SETUP_JSON, version))

        setup_content['version'] = version
        with open(FILEPATH_SETUP_JSON, 'w') as fil:
            # Write with indentation of two spaces and explicitly define separators to not have spaces at end of lines
            json.dump(setup_content, fil, indent=2, separators=(',', ': '))

        sys.exit(1)


@cli.command('toml')
def validate_pyproject():
    """
    Ensure that the version of reentry in setup.json and pyproject.toml are identical
    """
    reentry_requirement = None
    for requirement in get_setup_json()['install_requires']:
        if 'reentry' in requirement:
            reentry_requirement = requirement
            break

    if reentry_requirement is None:
        click.echo('Could not find the reentry requirement in {}'.format(FILEPATH_SETUP_JSON), err=True)
        sys.exit(1)

    try:
        with open(FILEPATH_TOML, 'r') as handle:
            toml_string = handle.read()
    except IOError as exception:
        click.echo('Could not read the required file: {}'.format(FILEPATH_TOML), err=True)
        sys.exit(1)

    try:
        parsed_toml = toml.loads(toml_string)
    except Exception as exception:  # pylint: disable=broad-except
        click.echo('Could not parse {}: {}'.format(FILEPATH_TOML, exception), err=True)
        sys.exit(1)

    try:
        pyproject_toml_requires = parsed_toml['build-system']['requires']
    except KeyError as exception:
        click.echo('Could not retrieve the build-system requires list from {}'.format(FILEPATH_TOML), err=True)
        sys.exit(1)

    if reentry_requirement not in pyproject_toml_requires:
        click.echo(
            'Reentry requirement from {} {} is not mirrored in {}'.format(
                FILEPATH_SETUP_JSON, reentry_requirement, FILEPATH_TOML
            ),
            err=True
        )
        sys.exit(1)


@cli.command('conda')
def update_environment_yml():
    """
    Updates environment.yml file for conda.
    """
    import yaml
    import re

    # needed for ordered dict, see
    # https://stackoverflow.com/a/52621703
    yaml.add_representer(
        OrderedDict,
        lambda self, data: yaml.representer.SafeRepresenter.represent_dict(self, data.items()),
        Dumper=yaml.SafeDumper
    )

    # fix incompatibilities between conda and pypi
    replacements = {'psycopg2-binary': 'psycopg2', 'graphviz': 'python-graphviz'}
    install_requires = get_setup_json()['install_requires']

    conda_requires = []
    for req in install_requires:
        # skip packages required for specific python versions
        # (environment.yml aims at the latest python version)
        if req.find('python_version') != -1:
            continue

        for (regex, replacement) in iter(replacements.items()):
            req = re.sub(regex, replacement, req)

        conda_requires.append(req)

    environment = OrderedDict([
        ('name', 'aiida'),
        ('channels', ['defaults', 'conda-forge', 'etetoolkit']),
        ('dependencies', conda_requires),
    ])

    environment_filename = 'environment.yml'
    file_path = os.path.join(ROOT_DIR, environment_filename)
    with open(file_path, 'w') as env_file:
        env_file.write('# Usage: conda env create -n myenvname -f environment.yml python=3.6\n')
        yaml.safe_dump(
            environment, env_file, explicit_start=True, default_flow_style=False, encoding='utf-8', allow_unicode=True
        )


if __name__ == '__main__':
    cli()  # pylint: disable=no-value-for-parameter



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
Generic tests that need the be specific to sqlalchemy
"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from six.moves import range

from aiida.backends.testbase import AiidaTestCase
from aiida.orm import Data


class TestComputer(AiidaTestCase):
    """
    Test the Computer class.
    """

    def test_deletion(self):
        from aiida.orm import CalcJobNode
        from aiida.common.exceptions import InvalidOperation
        import aiida.backends.sqlalchemy

        newcomputer = self.backend.computers.create(
            name='testdeletioncomputer',
            hostname='localhost',
            transport_type='local',
            scheduler_type='pbspro')
        newcomputer.store()

        # This should be possible, because nothing is using this computer
        self.backend.computers.delete(newcomputer.id)

        node = CalcJobNode()
        node.computer = self.computer
        node.set_option('resources', {'num_machines': 1, 'num_mpiprocs_per_machine': 1})
        node.store()

        session = aiida.backends.sqlalchemy.get_scoped_session()

        # This should fail, because there is at least a calculation
        # using this computer (the one created just above)
        try:
            session.begin_nested()
            with self.assertRaises(InvalidOperation):
                self.backend.computers.delete(self.computer.id)
        finally:
            session.rollback()


class TestGroupsSqla(AiidaTestCase):
    """
     Characterized functions
     """

    def setUp(self):
        from aiida.orm.implementation import sqlalchemy as sqla
        super(TestGroupsSqla, self).setUp()
        self.assertIsInstance(self.backend, sqla.backend.SqlaBackend)

    def test_query(self):
        """
        Test if queries are working
        """
        from aiida.common.exceptions import NotExistent, MultipleObjectsError

        backend = self.backend

        simple_user = backend.users.create('simple@ton.com')

        g1 = backend.groups.create(label='testquery1', user=simple_user).store()
        self.addCleanup(lambda: backend.groups.delete(g1.id))
        g2 = backend.groups.create(label='testquery2', user=simple_user).store()
        self.addCleanup(lambda: backend.groups.delete(g2.id))

        n1 = Data().store().backend_entity
        n2 = Data().store().backend_entity
        n3 = Data().store().backend_entity
        n4 = Data().store().backend_entity

        g1.add_nodes([n1, n2])
        g2.add_nodes([n1, n3])

        # NOTE: Here we pass type_string to query and get calls so that these calls don't
        # find the autogroups (otherwise the assertions will fail)
        newuser = backend.users.create(email='test@email.xx')
        g3 = backend.groups.create(label='testquery3', user=newuser).store()

        # I should find it
        g1copy = backend.groups.get(uuid=g1.uuid)
        self.assertEquals(g1.pk, g1copy.pk)

        # Try queries
        res = backend.groups.query(nodes=n4, type_string='')
        self.assertEquals([_.pk for _ in res], [])

        res = backend.groups.query(nodes=n1, type_string='')
        self.assertEquals([_.pk for _ in res], [_.pk for _ in [g1, g2]])

        res = backend.groups.query(nodes=n2, type_string='')
        self.assertEquals([_.pk for _ in res], [_.pk for _ in [g1]])

        # I try to use 'get' with zero or multiple results
        with self.assertRaises(NotExistent):
            backend.groups.get(nodes=n4, type_string='')
        with self.assertRaises(MultipleObjectsError):
            backend.groups.get(nodes=n1, type_string='')

        self.assertEquals(backend.groups.get(nodes=n2, type_string='').pk, g1.pk)

        # Query by user
        res = backend.groups.query(user=newuser, type_string='')
        self.assertSetEqual(set(_.pk for _ in res), set(_.pk for _ in [g3]))

        # Same query, but using a string (the username=email) instead of
        # a DbUser object
        res = backend.groups.query(user=newuser, type_string='')
        self.assertSetEqual(set(_.pk for _ in res), set(_.pk for _ in [g3]))

        res = backend.groups.query(user=simple_user, type_string='')

        self.assertSetEqual(set(_.pk for _ in res), set(_.pk for _ in [g1, g2]))


class TestGroupNoOrmSQLA(AiidaTestCase):
    """
    These tests check that the group node addition works ok when the skip_orm=True flag is used
    """

    def test_group_general(self):
        """
        General tests to verify that the group addition with the skip_orm=True flag
        work properly
        """
        backend = self.backend

        node_01 = Data().store().backend_entity
        node_02 = Data().store().backend_entity
        node_03 = Data().store().backend_entity
        node_04 = Data().store().backend_entity
        node_05 = Data().store().backend_entity
        nodes = [node_01, node_02, node_03, node_04, node_05]

        simple_user = backend.users.create('simple1@ton.com')
        group = backend.groups.create(label='test_adding_nodes', user=simple_user).store()
        # Single node in a list
        group.add_nodes([node_01], skip_orm=True)
        # List of nodes
        group.add_nodes([node_02, node_03], skip_orm=True)
        # Tuple of nodes
        group.add_nodes((node_04, node_05), skip_orm=True)

        # Check
        self.assertEqual(set(_.pk for _ in nodes), set(_.pk for _ in group.nodes))

        # Try to add a node that is already present: there should be no problem
        group.add_nodes([node_01], skip_orm=True)
        self.assertEqual(set(_.pk for _ in nodes), set(_.pk for _ in group.nodes))

    def test_group_batch_size(self):
        """
        Test that the group addition in batches works as expected.
        """
        from aiida.orm.groups import Group

        # Create 100 nodes
        nodes = []
        for _ in range(100):
            nodes.append(Data().store().backend_entity)

        # Add nodes to groups using different batch size. Check in the end the
        # correct addition.
        batch_sizes = (1, 3, 10, 1000)
        for batch_size in batch_sizes:
            group = Group(label='test_batches_' + str(batch_size)).store()
            group.backend_entity.add_nodes(nodes, skip_orm=True, batch_size=batch_size)
            self.assertEqual(set(_.pk for _ in nodes), set(_.pk for _ in group.nodes))



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Tests for `verdi user`."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from click.testing import CliRunner

from aiida.backends.testbase import AiidaTestCase
from aiida.cmdline.commands import cmd_user
from aiida import orm

user_1 = {
    'email': 'testuser1@localhost',
    'first_name': 'Max',
    'last_name': 'Mueller',
    'institution': 'Testing Instiute'
}
user_2 = {
    'email': 'testuser2@localhost',
    'first_name': 'Sabine',
    'last_name': 'Garching',
    'institution': 'Second testing instiute'
}


class TestVerdiUserCommand(AiidaTestCase):

    def setUp(self):
        super(TestVerdiUserCommand, self).setUp()

        created, user = orm.User.objects.get_or_create(email=user_1['email'])
        for key, value in user_1.items():
            if key != 'email':
                setattr(user, key, user_1[key])
        if created:
            orm.User(**user_1).store()
        self.cli_runner = CliRunner()

    def test_user_list(self):
        """Test `verdi user list`."""
        from aiida.cmdline.commands.cmd_user import user_list as list_user

        result = self.cli_runner.invoke(list_user, [], catch_exceptions=False)
        self.assertTrue(user_1['email'] in result.output)

    def test_user_create(self):
        """Create a new user with `verdi user configure`."""
        cli_options = [
            '--email', user_2['email'],
            '--first-name', user_2['first_name'],
            '--last-name', user_2['last_name'],
            '--institution', user_2['institution'],
        ]

        result = self.cli_runner.invoke(cmd_user.user_configure, cli_options, catch_exceptions=False)
        self.assertTrue(user_2['email'] in result.output)
        self.assertTrue('created' in result.output)
        self.assertTrue('updated' not in result.output)

        user_obj = orm.User.objects.get(email=user_2['email'])
        for key, val in user_2.items():
            self.assertEqual(val, getattr(user_obj, key))

    def test_user_update(self):
        """Reconfigure an existing user with `verdi user configure`."""
        email = user_1['email']

        cli_options = [
            '--email', user_1['email'],
            '--first-name', user_2['first_name'],
            '--last-name', user_2['last_name'],
            '--institution', user_2['institution'],
        ]

        result = self.cli_runner.invoke(cmd_user.user_configure, cli_options, catch_exceptions=False)
        self.assertTrue(email in result.output)
        self.assertTrue('updated' in result.output)
        self.assertTrue('created' not in result.output)

        # Check it's all been changed to user2's attributes except the email
        for key, value in user_2.items():
            if key != 'email':
                setattr(cmd_user, key, user_1[key])



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""`verdi user` command."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from functools import partial
import click

from aiida.cmdline.commands.cmd_verdi import verdi
from aiida.cmdline.params import arguments, options, types
from aiida.cmdline.utils import decorators, echo


def set_default_user(profile, user):
    """Set the user as the default user for the given profile.

    :param profile: the profile
    :param user: the user
    """
    from aiida.manage.configuration import get_config
    config = get_config()
    profile.default_user = user.email
    config.update_profile(profile)
    config.store()


def get_user_attribute_default(attribute, ctx):
    """Return the default value for the given attribute of the user passed in the context.

    :param attribute: attribute for which to get the current value
    :param ctx: click context which should contain the selected user
    :return: user attribute default value if set, or None
    """
    default = getattr(ctx.params['user'], attribute)

    # None or empty string means there is no default
    if not default:
        return None

    return default


@verdi.group('user')
def verdi_user():
    """Inspect and manage users."""


@verdi_user.command('list')
@decorators.with_dbenv()
def user_list():
    """Show a list of all users."""
    from aiida.orm import User

    default_user = User.objects.get_default()

    if default_user is None:
        echo.echo_warning('no default user has been configured')

    attributes = ['email', 'first_name', 'last_name']
    sort = lambda user: user.email
    highlight = lambda x: x.email == default_user.email if default_user else None
    echo.echo_formatted_list(User.objects.all(), attributes, sort=sort, highlight=highlight)


@verdi_user.command('configure')
@click.option(
    '--email',
    'user',
    prompt='User email',
    help='Email address that serves as the user name and a way to identify data created by it.',
    type=types.UserParamType(create=True),
    cls=options.interactive.InteractiveOption
)
@click.option(
    '--first-name',
    prompt='First name',
    help='First name of the user.',
    type=click.STRING,
    contextual_default=partial(get_user_attribute_default, 'first_name'),
    cls=options.interactive.InteractiveOption
)
@click.option(
    '--last-name',
    prompt='Last name',
    help='Last name of the user.',
    type=click.STRING,
    contextual_default=partial(get_user_attribute_default, 'last_name'),
    cls=options.interactive.InteractiveOption
)
@click.option(
    '--institution',
    prompt='Institution',
    help='Institution of the user.',
    type=click.STRING,
    contextual_default=partial(get_user_attribute_default, 'institution'),
    cls=options.interactive.InteractiveOption
)
@click.option(
    '--set-default',
    prompt='Set as default?',
    help='Set the user as the default user for the current profile.',
    is_flag=True,
    cls=options.interactive.InteractiveOption
)
@click.pass_context
@decorators.with_dbenv()
def user_configure(ctx, user, first_name, last_name, institution, set_default):
    """Configure a new or existing user.

    An e-mail address is used as the user name.
    """
    # pylint: disable=too-many-arguments
    if first_name is not None:
        user.first_name = first_name
    if last_name is not None:
        user.last_name = last_name
    if institution is not None:
        user.institution = institution

    action = 'updated' if user.is_stored else 'created'

    user.store()

    echo.echo_success('{} successfully {}'.format(user.email, action))

    if set_default:
        ctx.invoke(user_set_default, user=user)


@verdi_user.command('set-default')
@arguments.USER()
@click.pass_context
@decorators.with_dbenv()
def user_set_default(ctx, user):
    """Set a user as the default user for the profile."""
    set_default_user(ctx.obj.profile, user)
    echo.echo_success('set `{}` as the new default user for profile `{}`'.format(user.email, ctx.obj.profile.name))



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Test export file migration from export version 0.5 to 0.6"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from aiida.backends.general.migrations.calc_state import STATE_MAPPING
from aiida.backends.testbase import AiidaTestCase
from aiida.backends.tests.utils.archives import get_json_files
from aiida.tools.importexport.migration.utils import verify_metadata_version
from aiida.tools.importexport.migration.v05_to_v06 import migrate_v5_to_v6


class TestMigrateV05toV06(AiidaTestCase):
    """Test migration of export files from export version 0.5 to 0.6"""

    @classmethod
    def setUpClass(cls, *args, **kwargs):
        super(TestMigrateV05toV06, cls).setUpClass(*args, **kwargs)

        # Utility helpers
        cls.external_archive = {'filepath': 'archives', 'external_module': 'aiida-export-migration-tests'}
        cls.core_archive = {'filepath': 'export/migrate'}

    def test_migrate_v5_to_v6(self):
        """Test migration for file containing complete v0.5 era possibilities"""
        from aiida import get_version

        # Get metadata.json and data.json as dicts from v0.5 file archive
        metadata_v5, data_v5 = get_json_files('export_v0.5_simple.aiida', **self.core_archive)
        verify_metadata_version(metadata_v5, version='0.5')

        # Get metadata.json and data.json as dicts from v0.6 file archive
        metadata_v6, data_v6 = get_json_files('export_v0.6_simple.aiida', **self.core_archive)
        verify_metadata_version(metadata_v6, version='0.6')

        # Migrate to v0.6
        migrate_v5_to_v6(metadata_v5, data_v5)
        verify_metadata_version(metadata_v5, version='0.6')

        # Remove AiiDA version, since this may change irregardless of the migration function
        metadata_v5.pop('aiida_version')
        metadata_v6.pop('aiida_version')

        # Assert conversion message in `metadata.json` is correct and then remove it for later assertions
        self.maxDiff = None  # pylint: disable=invalid-name
        conversion_message = 'Converted from version 0.5 to 0.6 with AiiDA v{}'.format(get_version())
        self.assertEqual(
            metadata_v5.pop('conversion_info')[-1],
            conversion_message,
            msg='The conversion message after migration is wrong'
        )
        metadata_v6.pop('conversion_info')

        # Assert changes were performed correctly
        self.assertDictEqual(
            metadata_v5,
            metadata_v6,
            msg='After migration, metadata.json should equal intended metadata.json from archives'
        )
        self.assertDictEqual(
            data_v5, data_v6, msg='After migration, data.json should equal intended data.json from archives'
        )

    def test_migrate_v5_to_v6_calc_states(self):
        """Test the data migration of legacy `JobCalcState` attributes.

        This test has to use a local archive because the current archive from the `aiida-export-migration-tests`
        module does not include a `CalcJobNode` with a legacy `state` attribute.
        """
        # Get metadata.json and data.json as dicts from v0.5 file archive
        metadata, data = get_json_files('export_v0.5_simple.aiida', **self.core_archive)
        verify_metadata_version(metadata, version='0.5')

        calc_job_node_type = 'process.calculation.calcjob.CalcJobNode.'
        node_data = data['export_data'].get('Node', {})
        node_attributes = data['node_attributes']
        calc_jobs = {}
        for pk, values in node_data.items():
            if values['node_type'] == calc_job_node_type and 'state' in data['node_attributes'].get(pk, {}):
                calc_jobs[pk] = data['node_attributes'][pk]['state']

        # Migrate to v0.6
        migrate_v5_to_v6(metadata, data)
        verify_metadata_version(metadata, version='0.6')

        node_attributes = data['node_attributes']

        # The export archive contains a single `CalcJobNode` that had `state=FINISHED`.
        for pk, state in calc_jobs.items():

            attributes = node_attributes[pk]

            if STATE_MAPPING[state].exit_status is not None:
                self.assertEqual(attributes['exit_status'], STATE_MAPPING[state].exit_status)

            if STATE_MAPPING[state].process_state is not None:
                self.assertEqual(attributes['process_state'], STATE_MAPPING[state].process_state)

            if STATE_MAPPING[state].process_status is not None:
                self.assertEqual(attributes['process_status'], STATE_MAPPING[state].process_status)

            self.assertEqual(attributes['process_label'], 'Legacy JobCalculation')

    def test_migrate_v5_to_v6_datetime(self):
        """Test the data migration of serialized datetime objects.

        Datetime attributes were serialized into strings, by first converting to UTC and then printing with the format
        '%Y-%m-%dT%H:%M:%S.%f'. In the database migration, datetimes were serialized *including* timezone information.
        Here we test that the archive migration correctly reattaches the timezone information. The archive that we are
        using `export_v0.5_simple.aiida` contains a node with the attribute "scheduler_lastchecktime".
        """
        # Get metadata.json and data.json as dicts from v0.5 file archive
        metadata, data = get_json_files('export_v0.5_simple.aiida', **self.core_archive)
        verify_metadata_version(metadata, version='0.5')

        for key, values in data['node_attributes'].items():
            if 'scheduler_lastchecktime' not in values:
                continue

            serialized_original = values['scheduler_lastchecktime']
            msg = 'the serialized datetime before migration should not contain a plus: {}'.format(serialized_original)
            self.assertTrue('+' not in serialized_original, msg=msg)

            # Migrate to v0.6
            migrate_v5_to_v6(metadata, data)
            verify_metadata_version(metadata, version='0.6')

            serialized_migrated = data['node_attributes'][key]['scheduler_lastchecktime']
            self.assertEqual(serialized_migrated, serialized_original + '+00:00')
            break

        else:
            raise RuntimeError(
                'the archive `export_v0.5_simple.aiida` did not contain a node with the attribute '
                '`scheduler_lastchecktime` which is required for this test.'
            )

    def test_migrate_v5_to_v6_complete(self):
        """Test migration for file containing complete v0.5 era possibilities"""
        # Get metadata.json and data.json as dicts from v0.5 file archive
        metadata, data = get_json_files('export_v0.5_manual.aiida', **self.external_archive)
        verify_metadata_version(metadata, version='0.5')

        # Migrate to v0.6
        migrate_v5_to_v6(metadata, data)
        verify_metadata_version(metadata, version='0.6')

        self.maxDiff = None  # pylint: disable=invalid-name
        # Explicitly check that conversion dictionaries were removed
        illegal_data_dicts = {'node_attributes_conversion', 'node_extras_conversion'}
        for dict_ in illegal_data_dicts:
            self.assertNotIn(dict_, data, msg="dictionary '{}' should have been removed from data.json".format(dict_))



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
This module contains a set of unittest test classes that can be loaded from
the plugin.
Every transport plugin should be able to pass all of these common tests.
Plugin specific tests will be written in the plugin itself.
"""

# TODO : test for copy with pattern
# TODO : test for copy with/without patterns, overwriting folder
# TODO : test for exotic cases of copy with source = destination
# TODO : silly cases of copy/put/get from self to self

from __future__ import division
from __future__ import absolute_import
from __future__ import print_function
import io
import unittest

from six.moves import range


def get_all_custom_transports():
    """
    Autodiscover all custom transports defined in the variable
    plugin_transpors inside each test_* file in this folder.

    Therefore, do not move this function out of this file.

    :return: a dictionary of objects as defined in the various plugin_transport
      variables of the different files (the key is the module in which
      it was found)
    """
    import importlib
    import os

    modulename = __name__.rpartition('.')[0]
    this_full_fname = __file__
    thisdir, thisfname = os.path.split(this_full_fname)

    test_modules = [
        os.path.split(fname)[1][:-3] for fname in os.listdir(thisdir) if fname.endswith('.py') and fname.startswith('test_')
    ]

    # Remove this module: note that I should be careful because __file__, from
    # the second time on, is the pyc file rather than the py file
    thisbasename = os.path.splitext(thisfname)[0]
    try:
        test_modules.remove(thisbasename)
    except IndexError:
        print('Warning, this module ({}) was not found!'.format(thisbasename))

    all_custom_transports = {}
    for m in test_modules:
        module = importlib.import_module('.'.join([modulename, m]))
        custom_transport = module.__dict__.get('plugin_transport', None)
        if custom_transport is None:
            print('Define the plugin_transport variable inside the {} module!'.format(m))
        else:
            all_custom_transports[m] = custom_transport

    return all_custom_transports


def run_for_all_plugins(actual_test_method):
    """
    Decorator method that actually run the methods with an additional
    parameter (custom_transport), once for every custom_transport defined
    in the test_* files [except this one].
    """

    class CollectiveException(Exception):
        pass

    all_custom_transports = get_all_custom_transports()

    def test_all_plugins(self):
        """
        The wrapper function that calls the subfunction for each transport.
        """
        exceptions = []
        for tr_name, custom_transport in all_custom_transports.items():
            try:
                actual_test_method(self, custom_transport)
            except Exception as e:
                import traceback

                exceptions.append((e, traceback.format_exc(), tr_name))

        if exceptions:
            if all(isinstance(exc[0], AssertionError) for exc in exceptions):
                exception_to_raise = AssertionError
            else:
                exception_to_raise = CollectiveException

            messages = ['*** At least one test for a subplugin failed. See below ***', '']
            for exc in exceptions:
                messages.append("*** [For plugin {}]: Exception '{}': {}"
                                .format(exc[2], type(exc[0]).__name__, exc[0]))
                messages.append(exc[1])

            raise exception_to_raise('\n'.join(messages))

    return test_all_plugins


class TestBasicFunctionality(unittest.TestCase):
    """
    Tests to check basic functionality of transports.
    """

    @run_for_all_plugins
    def test_is_open(self, custom_transport):
        """Test that the is_open property works."""
        self.assertFalse(custom_transport.is_open)

        with custom_transport:
            self.assertTrue(custom_transport.is_open)

        self.assertFalse(custom_transport.is_open)


class TestDirectoryManipulation(unittest.TestCase):
    """
    Tests to check, create and delete folders.
    """

    @run_for_all_plugins
    def test_makedirs(self, custom_transport):
        """
        Verify the functioning of makedirs command
        """
        # Imports required later
        import random
        import string
        import os

        with custom_transport as t:
            location = t.normalize(os.path.join('/', 'tmp'))
            directory = 'temp_dir_test'
            t.chdir(location)

            self.assertEquals(location, t.getcwd())
            while t.isdir(directory):
                # I append a random letter/number until it is unique
                directory += random.choice(string.ascii_uppercase + string.digits)
            t.mkdir(directory)
            t.chdir(directory)

            # define folder structure
            dir_tree = os.path.join('1', '2')
            # I create the tree
            t.makedirs(dir_tree)
            # verify the existence
            self.assertTrue(t.isdir('1'))
            self.assertTrue(dir_tree)

            # try to recreate the same folder
            with self.assertRaises(OSError):
                t.makedirs(dir_tree)

            # recreate but with ignore flag
            t.makedirs(dir_tree, True)

            t.rmdir(dir_tree)
            t.rmdir('1')

            t.chdir('..')
            t.rmdir(directory)

    @run_for_all_plugins
    def test_rmtree(self, custom_transport):
        """
        Verify the functioning of rmtree command
        """
        # Imports required later
        import random
        import string
        import os

        with custom_transport as t:
            location = t.normalize(os.path.join('/', 'tmp'))
            directory = 'temp_dir_test'
            t.chdir(location)

            self.assertEquals(location, t.getcwd())
            while t.isdir(directory):
                # I append a random letter/number until it is unique
                directory += random.choice(string.ascii_uppercase + string.digits)
            t.mkdir(directory)
            t.chdir(directory)

            # define folder structure
            dir_tree = os.path.join('1', '2')
            # I create the tree
            t.makedirs(dir_tree)
            # remove it
            t.rmtree('1')
            # verify the removal
            self.assertFalse(t.isdir('1'))

            # also tests that it works with a single file
            # create file
            local_file_name = 'file.txt'
            text = u'Viva Verdi\n'
            with io.open(os.path.join(t.getcwd(), local_file_name), 'w', encoding='utf8') as fhandle:
                fhandle.write(text)
            # remove it
            t.rmtree(local_file_name)
            # verify the removal
            self.assertFalse(t.isfile(local_file_name))

            t.chdir('..')
            t.rmdir(directory)

    @run_for_all_plugins
    def test_listdir(self, custom_transport):
        """
        create directories, verify listdir, delete a folder with subfolders
        """
        # Imports required later
        import tempfile
        import random
        import string
        import os

        with custom_transport as trans:
            # We cannot use tempfile.mkdtemp because we're on a remote folder
            location = trans.normalize(os.path.join('/', 'tmp'))
            directory = 'temp_dir_test'
            trans.chdir(location)

            self.assertEquals(location, trans.getcwd())
            while trans.isdir(directory):
                # I append a random letter/number until it is unique
                directory += random.choice(string.ascii_uppercase + string.digits)
            trans.mkdir(directory)
            trans.chdir(directory)
            list_of_dir = ['1', '-f a&', 'as', 'a2', 'a4f']
            list_of_files = ['a', 'b']
            for this_dir in list_of_dir:
                trans.mkdir(this_dir)
            for fname in list_of_files:
                with tempfile.NamedTemporaryFile() as tmpf:
                    # Just put an empty file there at the right file name
                    trans.putfile(tmpf.name, fname)

            list_found = trans.listdir('.')

            self.assertTrue(sorted(list_found) == sorted(list_of_dir + list_of_files))

            self.assertTrue(sorted(trans.listdir('.', 'a*')), sorted(['as', 'a2', 'a4f']))
            self.assertTrue(sorted(trans.listdir('.', 'a?')), sorted(['as', 'a2']))
            self.assertTrue(sorted(trans.listdir('.', 'a[2-4]*')), sorted(['a2', 'a4f']))

            for this_dir in list_of_dir:
                trans.rmdir(this_dir)

            for this_file in list_of_files:
                trans.remove(this_file)

            trans.chdir('..')
            trans.rmdir(directory)

    @run_for_all_plugins
    def test_listdir_withattributes(self, custom_transport):
        """
        create directories, verify listdir_withattributes, delete a folder with subfolders
        """
        # Imports required later
        import tempfile
        import random
        import string
        import os

        def simplify_attributes(data):
            """
            Take data from listdir_withattributes and return a dictionary
            {fname: isdir}

            :param data: the output of listdir_withattributes
            :return: dictionary: the key is a filename, the value is True if it's a directory, False otherwise
            """
            return {_['name']: _['isdir'] for _ in data}

        with custom_transport as trans:
            # We cannot use tempfile.mkdtemp because we're on a remote folder
            location = trans.normalize(os.path.join('/', 'tmp'))
            directory = 'temp_dir_test'
            trans.chdir(location)

            self.assertEquals(location, trans.getcwd())
            while trans.isdir(directory):
                # I append a random letter/number until it is unique
                directory += random.choice(string.ascii_uppercase + string.digits)
            trans.mkdir(directory)
            trans.chdir(directory)
            list_of_dir = ['1', '-f a&', 'as', 'a2', 'a4f']
            list_of_files = ['a', 'b']
            for this_dir in list_of_dir:
                trans.mkdir(this_dir)
            for fname in list_of_files:
                with tempfile.NamedTemporaryFile() as tmpf:
                    # Just put an empty file there at the right file name
                    trans.putfile(tmpf.name, fname)

            comparison_list = {k: True for k in list_of_dir}
            for k in list_of_files:
                comparison_list[k] = False
            self.assertTrue(simplify_attributes(trans.listdir_withattributes('.')), comparison_list)

            self.assertTrue(
                simplify_attributes(trans.listdir_withattributes('.', 'a*')), {
                    'as': True,
                    'a2': True,
                    'a4f': True,
                    'a': False
                })
            self.assertTrue(simplify_attributes(trans.listdir_withattributes('.', 'a?')), {'as': True, 'a2': True})
            self.assertTrue(simplify_attributes(trans.listdir_withattributes('.', 'a[2-4]*')), {'a2': True, 'a4f': True})

            for this_dir in list_of_dir:
                trans.rmdir(this_dir)

            for this_file in list_of_files:
                trans.remove(this_file)

            trans.chdir('..')
            trans.rmdir(directory)

    @run_for_all_plugins
    def test_dir_creation_deletion(self, custom_transport):
        # Imports required later
        import random
        import string
        import os

        with custom_transport as t:
            location = t.normalize(os.path.join('/', 'tmp'))
            directory = 'temp_dir_test'
            t.chdir(location)

            self.assertEquals(location, t.getcwd())
            while t.isdir(directory):
                # I append a random letter/number until it is unique
                directory += random.choice(string.ascii_uppercase + string.digits)
            t.mkdir(directory)

            with self.assertRaises(OSError):
                # I create twice the same directory
                t.mkdir(directory)

            t.isdir(directory)
            self.assertFalse(t.isfile(directory))
            t.rmdir(directory)

    @run_for_all_plugins
    def test_dir_copy(self, custom_transport):
        """
        Verify if in the copy of a directory also the protection bits
        are carried over
        """
        # Imports required later
        import random
        import string
        import os

        with custom_transport as t:
            location = t.normalize(os.path.join('/', 'tmp'))
            directory = 'temp_dir_test'
            t.chdir(location)

            while t.isdir(directory):
                # I append a random letter/number until it is unique
                directory += random.choice(string.ascii_uppercase + string.digits)
            t.mkdir(directory)

            dest_directory = directory + '_copy'
            t.copy(directory, dest_directory)

            with self.assertRaises(ValueError):
                t.copy(directory, '')

            with self.assertRaises(ValueError):
                t.copy('', directory)

            t.rmdir(directory)
            t.rmdir(dest_directory)

    @run_for_all_plugins
    def test_dir_permissions_creation_modification(self, custom_transport):
        """
        verify if chmod raises IOError when trying to change bits on a
        non-existing folder
        """
        # Imports required later
        import random
        import string
        import os

        with custom_transport as t:
            location = t.normalize(os.path.join('/', 'tmp'))
            directory = 'temp_dir_test'
            t.chdir(location)

            while t.isdir(directory):
                # I append a random letter/number until it is unique
                directory += random.choice(string.ascii_uppercase + string.digits)

            # create directory with non default permissions
            t.mkdir(directory)

            # change permissions
            t.chmod(directory, 0o777)

            # test if the security bits have changed
            self.assertEquals(t.get_mode(directory), 0o777)

            # change permissions
            t.chmod(directory, 0o511)

            # test if the security bits have changed
            self.assertEquals(t.get_mode(directory), 0o511)

            # TODO : bug in paramiko. When changing the directory to very low \
            # I cannot set it back to higher permissions

            ## TODO: probably here we should then check for
            ## the new directory modes. To see if we want a higher
            ## level function to ask for the mode, or we just
            ## use get_attribute
            t.chdir(directory)

            # change permissions of an empty string, non existing folder.
            fake_dir = ''
            with self.assertRaises(IOError):
                t.chmod(fake_dir, 0o777)

            fake_dir = 'pippo'
            with self.assertRaises(IOError):
                # chmod to a non existing folder
                t.chmod(fake_dir, 0o777)

            t.chdir('..')
            t.rmdir(directory)

    @run_for_all_plugins
    def test_dir_reading_permissions(self, custom_transport):
        """
        Try to enter a directory with no read permissions.
        Verify that the cwd has not changed after failed try.
        """
        # Imports required later
        import random
        import string
        import os

        with custom_transport as t:
            location = t.normalize(os.path.join('/', 'tmp'))
            directory = 'temp_dir_test'
            t.chdir(location)

            while t.isdir(directory):
                # I append a random letter/number until it is unique
                directory += random.choice(string.ascii_uppercase + string.digits)

            # create directory with non default permissions
            t.mkdir(directory)

            # change permissions to low ones
            t.chmod(directory, 0)

            # test if the security bits have changed
            self.assertEquals(t.get_mode(directory), 0)

            old_cwd = t.getcwd()

            with self.assertRaises(IOError):
                t.chdir(directory)

            new_cwd = t.getcwd()

            self.assertEquals(old_cwd, new_cwd)

            # TODO : the test leaves a directory even if it is successful
            #        The bug is in paramiko. After lowering the permissions,
            #        I cannot restore them to higher values
            #t.rmdir(directory)

    @run_for_all_plugins
    def test_isfile_isdir_to_empty_string(self, custom_transport):
        """
        I check that isdir or isfile return False when executed on an
        empty string
        """
        import os

        with custom_transport as t:
            location = t.normalize(os.path.join('/', 'tmp'))
            t.chdir(location)
            self.assertFalse(t.isdir(''))
            self.assertFalse(t.isfile(''))

    @run_for_all_plugins
    def test_isfile_isdir_to_non_existing_string(self, custom_transport):
        """
        I check that isdir or isfile return False when executed on an
        empty string
        """
        import os

        with custom_transport as t:
            location = t.normalize(os.path.join('/', 'tmp'))
            t.chdir(location)
            fake_folder = 'pippo'
            self.assertFalse(t.isfile(fake_folder))
            self.assertFalse(t.isdir(fake_folder))
            with self.assertRaises(IOError):
                t.chdir(fake_folder)

    @run_for_all_plugins
    def test_chdir_to_empty_string(self, custom_transport):
        """
        I check that if I pass an empty string to chdir, the cwd does
        not change (this is a paramiko default behavior), but getcwd()
        is still correctly defined.
        """
        import os

        with custom_transport as t:
            new_dir = t.normalize(os.path.join('/', 'tmp'))
            t.chdir(new_dir)
            t.chdir('')
            self.assertEquals(new_dir, t.getcwd())


class TestPutGetFile(unittest.TestCase):
    """
    Test to verify whether the put and get functions behave correctly on files.
    1) they work
    2) they need abs paths where necessary, i.e. for local paths
    3) they reject empty strings
    """

    @run_for_all_plugins
    def test_put_and_get(self, custom_transport):
        import os
        import random
        import string

        local_dir = os.path.join('/', 'tmp')
        remote_dir = local_dir
        directory = 'tmp_try'

        with custom_transport as t:
            t.chdir(remote_dir)
            while t.isdir(directory):
                # I append a random letter/number until it is unique
                directory += random.choice(string.ascii_uppercase + string.digits)

            t.mkdir(directory)
            t.chdir(directory)

            local_file_name = os.path.join(local_dir, directory, 'file.txt')
            remote_file_name = 'file_remote.txt'
            retrieved_file_name = os.path.join(local_dir, directory, 'file_retrieved.txt')

            text = u'Viva Verdi\n'
            with io.open(local_file_name, 'w', encoding='utf8') as fhandle:
                fhandle.write(text)

            # here use full path in src and dst
            t.put(local_file_name, remote_file_name)
            t.get(remote_file_name, retrieved_file_name)
            t.putfile(local_file_name, remote_file_name)
            t.getfile(remote_file_name, retrieved_file_name)

            list_of_files = t.listdir('.')
            # it is False because local_file_name has the full path,
            # while list_of_files has not
            self.assertFalse(local_file_name in list_of_files)
            self.assertTrue(remote_file_name in list_of_files)
            self.assertFalse(retrieved_file_name in list_of_files)

            os.remove(local_file_name)
            t.remove(remote_file_name)
            os.remove(retrieved_file_name)

            t.chdir('..')
            t.rmdir(directory)

    @run_for_all_plugins
    def test_put_get_abs_path(self, custom_transport):
        """
        test of exception for non existing files and abs path
        """
        import os
        import random
        import string

        local_dir = os.path.join('/', 'tmp')
        remote_dir = local_dir
        directory = 'tmp_try'

        with custom_transport as t:
            t.chdir(remote_dir)
            while t.isdir(directory):
                # I append a random letter/number until it is unique
                directory += random.choice(string.ascii_uppercase + string.digits)

            t.mkdir(directory)
            t.chdir(directory)

            partial_file_name = 'file.txt'
            local_file_name = os.path.join(local_dir, directory, 'file.txt')
            remote_file_name = 'file_remote.txt'
            retrieved_file_name = os.path.join(local_dir, directory, 'file_retrieved.txt')

            fhandle = io.open(local_file_name, 'w', encoding='utf8')
            fhandle.close()

            # partial_file_name is not an abs path
            with self.assertRaises(ValueError):
                t.put(partial_file_name, remote_file_name)
            with self.assertRaises(ValueError):
                t.putfile(partial_file_name, remote_file_name)

            # retrieved_file_name does not exist
            with self.assertRaises(OSError):
                t.put(retrieved_file_name, remote_file_name)
            with self.assertRaises(OSError):
                t.putfile(retrieved_file_name, remote_file_name)

            # remote_file_name does not exist
            with self.assertRaises(IOError):
                t.get(remote_file_name, retrieved_file_name)
            with self.assertRaises(IOError):
                t.getfile(remote_file_name, retrieved_file_name)

            t.put(local_file_name, remote_file_name)
            t.putfile(local_file_name, remote_file_name)

            # local filename is not an abs path
            with self.assertRaises(ValueError):
                t.get(remote_file_name, 'delete_me.txt')
            with self.assertRaises(ValueError):
                t.getfile(remote_file_name, 'delete_me.txt')

            t.remove(remote_file_name)
            os.remove(local_file_name)

            t.chdir('..')
            t.rmdir(directory)

    @run_for_all_plugins
    def test_put_get_empty_string(self, custom_transport):
        """
        test of exception put/get of empty strings
        """
        # TODO : verify the correctness of \n at the end of a file
        import os
        import random
        import string

        local_dir = os.path.join('/', 'tmp')
        remote_dir = local_dir
        directory = 'tmp_try'

        with custom_transport as t:
            t.chdir(remote_dir)
            while t.isdir(directory):
                # I append a random letter/number until it is unique
                directory += random.choice(string.ascii_uppercase + string.digits)

            t.mkdir(directory)
            t.chdir(directory)

            local_file_name = os.path.join(local_dir, directory, 'file_local.txt')
            remote_file_name = 'file_remote.txt'
            retrieved_file_name = os.path.join(local_dir, directory, 'file_retrieved.txt')

            text = u'Viva Verdi\n'
            with io.open(local_file_name, 'w', encoding='utf8') as fhandle:
                fhandle.write(text)

            # localpath is an empty string
            # ValueError because it is not an abs path
            with self.assertRaises(ValueError):
                t.put('', remote_file_name)
            with self.assertRaises(ValueError):
                t.putfile('', remote_file_name)

            # remote path is an empty string
            with self.assertRaises(IOError):
                t.put(local_file_name, '')
            with self.assertRaises(IOError):
                t.putfile(local_file_name, '')

            t.put(local_file_name, remote_file_name)
            # overwrite the remote_file_name
            t.putfile(local_file_name, remote_file_name)

            # remote path is an empty string
            with self.assertRaises(IOError):
                t.get('', retrieved_file_name)
            with self.assertRaises(IOError):
                t.getfile('', retrieved_file_name)

            # local path is an empty string
            # ValueError because it is not an abs path
            with self.assertRaises(ValueError):
                t.get(remote_file_name, '')
            with self.assertRaises(ValueError):
                t.getfile(remote_file_name, '')

            # TODO : get doesn't retrieve empty files.
            # Is it what we want?
            t.get(remote_file_name, retrieved_file_name)
            # overwrite retrieved_file_name
            t.getfile(remote_file_name, retrieved_file_name)

            os.remove(local_file_name)
            t.remove(remote_file_name)
            # If it couldn't end the copy, it leaves what he did on
            # local file
            self.assertTrue('file_retrieved.txt' in t.listdir('.'))
            os.remove(retrieved_file_name)

            t.chdir('..')
            t.rmdir(directory)


class TestPutGetTree(unittest.TestCase):
    """
    Test to verify whether the put and get functions behave correctly on folders.
    1) they work
    2) they need abs paths where necessary, i.e. for local paths
    3) they reject empty strings
    """

    @run_for_all_plugins
    def test_put_and_get(self, custom_transport):
        import os
        import random
        import string

        local_dir = os.path.join('/', 'tmp')
        remote_dir = local_dir
        directory = 'tmp_try'

        with custom_transport as t:

            t.chdir(remote_dir)

            while os.path.exists(os.path.join(local_dir, directory)):
                # I append a random letter/number until it is unique
                directory += random.choice(string.ascii_uppercase + string.digits)

            local_subfolder = os.path.join(local_dir, directory, 'tmp1')
            remote_subfolder = 'tmp2'
            retrieved_subfolder = os.path.join(local_dir, directory, 'tmp3')

            os.mkdir(os.path.join(local_dir, directory))
            os.mkdir(os.path.join(local_dir, directory, local_subfolder))

            t.chdir(directory)

            local_file_name = os.path.join(local_subfolder, 'file.txt')

            text = u'Viva Verdi\n'
            with io.open(local_file_name, 'w', encoding='utf8') as fhandle:
                fhandle.write(text)

            # here use full path in src and dst
            for i in range(2):
                if i == 0:
                    t.put(local_subfolder, remote_subfolder)
                    t.get(remote_subfolder, retrieved_subfolder)
                else:
                    t.puttree(local_subfolder, remote_subfolder)
                    t.gettree(remote_subfolder, retrieved_subfolder)

                # Here I am mixing the local with the remote fold
                list_of_dirs = t.listdir('.')
                # # it is False because local_file_name has the full path,
                # # while list_of_files has not
                self.assertFalse(local_subfolder in list_of_dirs)
                self.assertTrue(remote_subfolder in list_of_dirs)
                self.assertFalse(retrieved_subfolder in list_of_dirs)
                self.assertTrue('tmp1' in list_of_dirs)
                self.assertTrue('tmp3' in list_of_dirs)

                list_pushed_file = t.listdir('tmp2')
                list_retrieved_file = t.listdir('tmp3')
                self.assertTrue('file.txt' in list_pushed_file)
                self.assertTrue('file.txt' in list_retrieved_file)

            import shutil

            shutil.rmtree(local_subfolder)
            shutil.rmtree(retrieved_subfolder)
            t.rmtree(remote_subfolder)

            t.chdir('..')
            t.rmdir(directory)

    @run_for_all_plugins
    def test_put_and_get_overwrite(self, custom_transport):
        import os, shutil
        import random
        import string

        local_dir = os.path.join('/', 'tmp')
        remote_dir = local_dir
        directory = 'tmp_try'

        with custom_transport as t:
            t.chdir(remote_dir)

            while os.path.exists(os.path.join(local_dir, directory)):
                # I append a random letter/number until it is unique
                directory += random.choice(string.ascii_uppercase + string.digits)

            local_subfolder = os.path.join(local_dir, directory, 'tmp1')
            remote_subfolder = 'tmp2'
            retrieved_subfolder = os.path.join(local_dir, directory, 'tmp3')

            os.mkdir(os.path.join(local_dir, directory))
            os.mkdir(os.path.join(local_dir, directory, local_subfolder))

            t.chdir(directory)

            local_file_name = os.path.join(local_subfolder, 'file.txt')

            text = u'Viva Verdi\n'
            with io.open(local_file_name, 'w', encoding='utf8') as fhandle:
                fhandle.write(text)

            t.put(local_subfolder, remote_subfolder)
            t.get(remote_subfolder, retrieved_subfolder)

            # by defaults rewrite everything
            t.put(local_subfolder, remote_subfolder)
            t.get(remote_subfolder, retrieved_subfolder)

            with self.assertRaises(OSError):
                t.put(local_subfolder, remote_subfolder, overwrite=False)
            with self.assertRaises(OSError):
                t.get(remote_subfolder, retrieved_subfolder, overwrite=False)
            with self.assertRaises(OSError):
                t.puttree(local_subfolder, remote_subfolder, overwrite=False)
            with self.assertRaises(OSError):
                t.gettree(remote_subfolder, retrieved_subfolder, overwrite=False)

            shutil.rmtree(local_subfolder)
            shutil.rmtree(retrieved_subfolder)
            t.rmtree(remote_subfolder)
            # t.rmtree(remote_subfolder)
            # here I am mixing inevitably the local and the remote folder
            t.chdir('..')
            t.rmtree(directory)

    @run_for_all_plugins
    def test_copy(self, custom_transport):
        import os
        import random
        import string

        local_dir = os.path.join('/', 'tmp')
        remote_dir = local_dir
        directory = 'tmp_try'

        with custom_transport as t:
            t.chdir(remote_dir)

            while os.path.exists(os.path.join(local_dir, directory)):
                # I append a random letter/number until it is unique
                directory += random.choice(string.ascii_uppercase + string.digits)

            t.mkdir(directory)
            t.chdir(directory)

            local_base_dir = os.path.join(local_dir, directory, 'local')
            os.mkdir(local_base_dir)

            # first test put: I create three files in local
            file_1 = os.path.join(local_base_dir, 'a.txt')
            file_2 = os.path.join(local_base_dir, 'b.tmp')
            file_3 = os.path.join(local_base_dir, 'c.txt')
            text = u'Viva Verdi\n'
            for filename in [file_1, file_2, file_3]:
                with io.open(filename, 'w', encoding='utf8') as fhandle:
                    fhandle.write(text)

            # first test the copy. Copy of two files matching patterns, into a folder
            t.copy(os.path.join('local', '*.txt'), '.')
            self.assertEquals(set(['a.txt', 'c.txt', 'local']), set(t.listdir('.')))
            t.remove('a.txt')
            t.remove('c.txt')
            # second test copy. Copy of two folders
            t.copy('local', 'prova')
            self.assertEquals(set(['prova', 'local']), set(t.listdir('.')))
            self.assertEquals(set(['a.txt', 'b.tmp', 'c.txt']), set(t.listdir('prova')))
            t.rmtree('prova')
            # third test copy. Can copy one file into a new file
            t.copy(os.path.join('local', '*.tmp'), 'prova')
            self.assertEquals(set(['prova', 'local']), set(t.listdir('.')))
            t.remove('prova')
            # fourth test copy: can't copy more than one file on the same file,
            # i.e., the destination should be a folder
            with self.assertRaises(OSError):
                t.copy(os.path.join('local', '*.txt'), 'prova')
            # fifth test, copying one file into a folder
            t.mkdir('prova')
            t.copy(os.path.join('local', 'a.txt'), 'prova')
            self.assertEquals(set(t.listdir('prova')), set(['a.txt']))
            t.rmtree('prova')
            # sixth test, copying one file into a file
            t.copy(os.path.join('local', 'a.txt'), 'prova')
            self.assertTrue(t.isfile('prova'))
            t.remove('prova')
            # copy of folder into an existing folder
            #NOTE: the command cp has a different behavior on Mac vs Ubuntu
            #tests performed locally on a Mac may result in a failure.
            t.mkdir('prova')
            t.copy('local', 'prova')
            self.assertEquals(set(['local']), set(t.listdir('prova')))
            self.assertEquals(set(['a.txt', 'b.tmp', 'c.txt']), set(t.listdir(os.path.join('prova', 'local'))))
            t.rmtree('prova')
            # exit
            t.chdir('..')
            t.rmtree(directory)

    @run_for_all_plugins
    def test_put(self, custom_transport):
        # exactly the same tests of copy, just with the put function
        # and therefore the local path must be absolute
        import os
        import random
        import string

        local_dir = os.path.join('/', 'tmp')
        remote_dir = local_dir
        directory = 'tmp_try'

        with custom_transport as t:
            t.chdir(remote_dir)

            while os.path.exists(os.path.join(local_dir, directory)):
                # I append a random letter/number until it is unique
                directory += random.choice(string.ascii_uppercase + string.digits)

            t.mkdir(directory)
            t.chdir(directory)

            local_base_dir = os.path.join(local_dir, directory, 'local')
            os.mkdir(local_base_dir)

            # first test put: I create three files in local
            file_1 = os.path.join(local_base_dir, 'a.txt')
            file_2 = os.path.join(local_base_dir, 'b.tmp')
            file_3 = os.path.join(local_base_dir, 'c.txt')
            text = u'Viva Verdi\n'
            for filename in [file_1, file_2, file_3]:
                with io.open(filename, 'w', encoding='utf8') as fhandle:
                    fhandle.write(text)

            # first test put. Copy of two files matching patterns, into a folder
            t.put(os.path.join(local_base_dir, '*.txt'), '.')
            self.assertEquals(set(['a.txt', 'c.txt', 'local']), set(t.listdir('.')))
            t.remove('a.txt')
            t.remove('c.txt')
            # second. Copy of folder into a non existing folder
            t.put(local_base_dir, 'prova')
            self.assertEquals(set(['prova', 'local']), set(t.listdir('.')))
            self.assertEquals(set(['a.txt', 'b.tmp', 'c.txt']), set(t.listdir('prova')))
            t.rmtree('prova')
            # third. copy of folder into an existing folder
            t.mkdir('prova')
            t.put(local_base_dir, 'prova')
            self.assertEquals(set(['prova', 'local']), set(t.listdir('.')))
            self.assertEquals(set(['local']), set(t.listdir('prova')))
            self.assertEquals(set(['a.txt', 'b.tmp', 'c.txt']), set(t.listdir(os.path.join('prova', 'local'))))
            t.rmtree('prova')
            # third test copy. Can copy one file into a new file
            t.put(os.path.join(local_base_dir, '*.tmp'), 'prova')
            self.assertEquals(set(['prova', 'local']), set(t.listdir('.')))
            t.remove('prova')
            # fourth test copy: can't copy more than one file on the same file,
            # i.e., the destination should be a folder
            with self.assertRaises(OSError):
                t.put(os.path.join(local_base_dir, '*.txt'), 'prova')
            # copy of folder into file
            with io.open(os.path.join(local_dir, directory, 'existing.txt'), 'w', encoding='utf8') as fhandle:
                fhandle.write(text)
            with self.assertRaises(OSError):
                t.put(os.path.join(local_base_dir), 'existing.txt')
            t.remove('existing.txt')
            # fifth test, copying one file into a folder
            t.mkdir('prova')
            t.put(os.path.join(local_base_dir, 'a.txt'), 'prova')
            self.assertEquals(set(t.listdir('prova')), set(['a.txt']))
            t.rmtree('prova')
            # sixth test, copying one file into a file
            t.put(os.path.join(local_base_dir, 'a.txt'), 'prova')
            self.assertTrue(t.isfile('prova'))
            t.remove('prova')

            # exit
            t.chdir('..')
            t.rmtree(directory)

    @run_for_all_plugins
    def test_get(self, custom_transport):
        # exactly the same tests of copy, just with the put function
        # and therefore the local path must be absolute
        import os
        import random
        import string, shutil

        local_dir = os.path.join('/', 'tmp')
        remote_dir = local_dir
        directory = 'tmp_try'

        with custom_transport as t:
            t.chdir(remote_dir)

            while os.path.exists(os.path.join(local_dir, directory)):
                # I append a random letter/number until it is unique
                directory += random.choice(string.ascii_uppercase + string.digits)

            t.mkdir(directory)
            t.chdir(directory)

            local_base_dir = os.path.join(local_dir, directory, 'local')
            local_destination = os.path.join(local_dir, directory)
            os.mkdir(local_base_dir)

            # first test put: I create three files in local
            file_1 = os.path.join(local_base_dir, 'a.txt')
            file_2 = os.path.join(local_base_dir, 'b.tmp')
            file_3 = os.path.join(local_base_dir, 'c.txt')
            text = u'Viva Verdi\n'
            for filename in [file_1, file_2, file_3]:
                with io.open(filename, 'w', encoding='utf8') as fhandle:
                    fhandle.write(text)

            # first test put. Copy of two files matching patterns, into a folder
            t.get(os.path.join('local', '*.txt'), local_destination)
            self.assertEquals(set(['a.txt', 'c.txt', 'local']), set(os.listdir(local_destination)))
            os.remove(os.path.join(local_destination, 'a.txt'))
            os.remove(os.path.join(local_destination, 'c.txt'))
            # second. Copy of folder into a non existing folder
            t.get('local', os.path.join(local_destination, 'prova'))
            self.assertEquals(set(['prova', 'local']), set(os.listdir(local_destination)))
            self.assertEquals(
                set(['a.txt', 'b.tmp', 'c.txt']), set(os.listdir(os.path.join(local_destination, 'prova'))))
            shutil.rmtree(os.path.join(local_destination, 'prova'))
            # third. copy of folder into an existing folder
            os.mkdir(os.path.join(local_destination, 'prova'))
            t.get('local', os.path.join(local_destination, 'prova'))
            self.assertEquals(set(['prova', 'local']), set(os.listdir(local_destination)))
            self.assertEquals(set(['local']), set(os.listdir(os.path.join(local_destination, 'prova'))))
            self.assertEquals(
                set(['a.txt', 'b.tmp', 'c.txt']), set(os.listdir(os.path.join(local_destination, 'prova', 'local'))))
            shutil.rmtree(os.path.join(local_destination, 'prova'))
            # third test copy. Can copy one file into a new file
            t.get(os.path.join('local', '*.tmp'), os.path.join(local_destination, 'prova'))
            self.assertEquals(set(['prova', 'local']), set(os.listdir(local_destination)))
            os.remove(os.path.join(local_destination, 'prova'))
            # fourth test copy: can't copy more than one file on the same file,
            # i.e., the destination should be a folder
            with self.assertRaises(OSError):
                t.get(os.path.join('local', '*.txt'), os.path.join(local_destination, 'prova'))
            # copy of folder into file
            with io.open(os.path.join(local_destination, 'existing.txt'), 'w', encoding='utf8') as fhandle:
                fhandle.write(text)
            with self.assertRaises(OSError):
                t.get('local', os.path.join(local_destination, 'existing.txt'))
            os.remove(os.path.join(local_destination, 'existing.txt'))
            # fifth test, copying one file into a folder
            os.mkdir(os.path.join(local_destination, 'prova'))
            t.get(os.path.join('local', 'a.txt'), os.path.join(local_destination, 'prova'))
            self.assertEquals(set(os.listdir(os.path.join(local_destination, 'prova'))), set(['a.txt']))
            shutil.rmtree(os.path.join(local_destination, 'prova'))
            # sixth test, copying one file into a file
            t.get(os.path.join('local', 'a.txt'), os.path.join(local_destination, 'prova'))
            self.assertTrue(os.path.isfile(os.path.join(local_destination, 'prova')))
            os.remove(os.path.join(local_destination, 'prova'))

            # exit
            t.chdir('..')
            t.rmtree(directory)

    @run_for_all_plugins
    def test_put_get_abs_path(self, custom_transport):
        """
        test of exception for non existing files and abs path
        """
        import os
        import random
        import string

        local_dir = os.path.join('/', 'tmp')
        remote_dir = local_dir
        directory = 'tmp_try'

        with custom_transport as t:
            t.chdir(remote_dir)

            while os.path.exists(os.path.join(local_dir, directory)):
                # I append a random letter/number until it is unique
                directory += random.choice(string.ascii_uppercase + string.digits)

            local_subfolder = os.path.join(local_dir, directory, 'tmp1')
            remote_subfolder = 'tmp2'
            retrieved_subfolder = os.path.join(local_dir, directory, 'tmp3')

            os.mkdir(os.path.join(local_dir, directory))
            os.mkdir(os.path.join(local_dir, directory, local_subfolder))

            t.chdir(directory)
            local_file_name = os.path.join(local_subfolder, 'file.txt')

            fhandle = io.open(local_file_name, 'w', encoding='utf8')
            fhandle.close()

            # 'tmp1' is not an abs path
            with self.assertRaises(ValueError):
                t.put('tmp1', remote_subfolder)
            with self.assertRaises(ValueError):
                t.putfile('tmp1', remote_subfolder)
            with self.assertRaises(ValueError):
                t.puttree('tmp1', remote_subfolder)

            # 'tmp3' does not exist
            with self.assertRaises(OSError):
                t.put(retrieved_subfolder, remote_subfolder)
            with self.assertRaises(OSError):
                t.putfile(retrieved_subfolder, remote_subfolder)
            with self.assertRaises(OSError):
                t.puttree(retrieved_subfolder, remote_subfolder)

            # remote_file_name does not exist
            with self.assertRaises(IOError):
                t.get('non_existing', retrieved_subfolder)
            with self.assertRaises(IOError):
                t.getfile('non_existing', retrieved_subfolder)
            with self.assertRaises(IOError):
                t.gettree('non_existing', retrieved_subfolder)

            t.put(local_subfolder, remote_subfolder)

            # local filename is not an abs path
            with self.assertRaises(ValueError):
                t.get(remote_subfolder, 'delete_me_tree')
            with self.assertRaises(ValueError):
                t.getfile(remote_subfolder, 'delete_me_tree')
            with self.assertRaises(ValueError):
                t.gettree(remote_subfolder, 'delete_me_tree')

            os.remove(os.path.join(local_subfolder, 'file.txt'))
            os.rmdir(local_subfolder)
            t.rmtree(remote_subfolder)

            t.chdir('..')
            t.rmdir(directory)

    @run_for_all_plugins
    def test_put_get_empty_string(self, custom_transport):
        """
        test of exception put/get of empty strings
        """
        # TODO : verify the correctness of \n at the end of a file
        import os
        import random
        import string

        local_dir = os.path.join('/', 'tmp')
        remote_dir = local_dir
        directory = 'tmp_try'

        with custom_transport as t:
            t.chdir(remote_dir)

            while os.path.exists(os.path.join(local_dir, directory)):
                # I append a random letter/number until it is unique
                directory += random.choice(string.ascii_uppercase + string.digits)

            local_subfolder = os.path.join(local_dir, directory, 'tmp1')
            remote_subfolder = 'tmp2'
            retrieved_subfolder = os.path.join(local_dir, directory, 'tmp3')

            os.mkdir(os.path.join(local_dir, directory))
            os.mkdir(os.path.join(local_dir, directory, local_subfolder))

            t.chdir(directory)
            local_file_name = os.path.join(local_subfolder, 'file.txt')

            text = u'Viva Verdi\n'
            with io.open(local_file_name, 'w', encoding='utf8') as fhandle:
                fhandle.write(text)

            # localpath is an empty string
            # ValueError because it is not an abs path
            with self.assertRaises(ValueError):
                t.puttree('', remote_subfolder)

            # remote path is an empty string
            with self.assertRaises(IOError):
                t.puttree(local_subfolder, '')

            t.puttree(local_subfolder, remote_subfolder)

            # remote path is an empty string
            with self.assertRaises(IOError):
                t.gettree('', retrieved_subfolder)

            # local path is an empty string
            # ValueError because it is not an abs path
            with self.assertRaises(ValueError):
                t.gettree(remote_subfolder, '')

            # TODO : get doesn't retrieve empty files.
            # Is it what we want?
            t.gettree(remote_subfolder, retrieved_subfolder)

            os.remove(os.path.join(local_subfolder, 'file.txt'))
            os.rmdir(local_subfolder)
            t.remove(os.path.join(remote_subfolder, 'file.txt'))
            t.rmdir(remote_subfolder)
            # If it couldn't end the copy, it leaves what he did on local file
            # here I am mixing local with remote
            self.assertTrue('file.txt' in t.listdir('tmp3'))
            os.remove(os.path.join(retrieved_subfolder, 'file.txt'))
            os.rmdir(retrieved_subfolder)

            t.chdir('..')
            t.rmdir(directory)


class TestExecuteCommandWait(unittest.TestCase):
    """
    Test some simple command executions and stdin/stdout management.

    It also checks for escaping of the folder names.
    """

    @run_for_all_plugins
    def test_exec_pwd(self, custom_transport):
        """
        I create a strange subfolder with a complicated name and
        then see if I can run pwd. This also checks the correct
        escaping of funny characters, both in the directory
        creation (which should be done by paramiko) and in the command
        execution (done in this module, in the _exec_command_internal function).
        """
        import os

        # Start value
        delete_at_end = False

        with custom_transport as t:

            # To compare with: getcwd uses the normalized ('realpath') path
            location = t.normalize('/tmp')
            subfolder = """_'s f"#"""  # A folder with characters to escape
            subfolder_fullpath = os.path.join(location, subfolder)

            t.chdir(location)
            if not t.isdir(subfolder):
                # Since I created the folder, I will remember to
                # delete it at the end of this test
                delete_at_end = True
                t.mkdir(subfolder)

            self.assertTrue(t.isdir(subfolder))
            t.chdir(subfolder)

            self.assertEquals(subfolder_fullpath, t.getcwd())
            retcode, stdout, stderr = t.exec_command_wait('pwd')
            self.assertEquals(retcode, 0)
            # I have to strip it because 'pwd' returns a trailing \n
            self.assertEquals(stdout.strip(), subfolder_fullpath)
            self.assertEquals(stderr, '')

            if delete_at_end:
                t.chdir(location)
                t.rmdir(subfolder)

    @run_for_all_plugins
    def test_exec_with_stdin_string(self, custom_transport):
        test_string = str('some_test String')
        with custom_transport as t:
            retcode, stdout, stderr = t.exec_command_wait('cat', stdin=test_string)
            self.assertEquals(retcode, 0)
            self.assertEquals(stdout, test_string)
            self.assertEquals(stderr, '')

    @run_for_all_plugins
    def test_exec_with_stdin_unicode(self, custom_transport):
        test_string = u'some_test String'
        with custom_transport as t:
            retcode, stdout, stderr = t.exec_command_wait('cat', stdin=test_string)
            self.assertEquals(retcode, 0)
            self.assertEquals(stdout, test_string)
            self.assertEquals(stderr, '')

    @run_for_all_plugins
    def test_exec_with_stdin_filelike(self, custom_transport):
        from six.moves import cStringIO as StringIO

        test_string = 'some_test String'
        stdin = StringIO(test_string)
        with custom_transport as t:
            retcode, stdout, stderr = t.exec_command_wait('cat', stdin=stdin)
            self.assertEquals(retcode, 0)
            self.assertEquals(stdout, test_string)
            self.assertEquals(stderr, '')

    @run_for_all_plugins
    def test_exec_with_wrong_stdin(self, custom_transport):
        # I pass a number
        with custom_transport as t:
            with self.assertRaises(ValueError):
                t.exec_command_wait('cat', stdin=1)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""ORM Log tests"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import logging

from six.moves import range

from aiida import orm
from aiida.backends.testbase import AiidaTestCase
from aiida.common import exceptions
from aiida.common.log import LOG_LEVEL_REPORT
from aiida.common.timezone import now
from aiida.orm import Log


class TestBackendLog(AiidaTestCase):
    """Test the Log entity"""

    def setUp(self):
        super(TestBackendLog, self).setUp()
        self.log_record = {
            'time': now(),
            'loggername': 'loggername',
            'levelname': logging.getLevelName(LOG_LEVEL_REPORT),
            'dbnode_id': None,
            'message': 'This is a template record message',
            'metadata': {
                'content': 'test'
            },
        }

    def tearDown(self):
        """
        Delete all the created log entries
        """
        super(TestBackendLog, self).tearDown()
        Log.objects.delete_all()

    def create_log(self):
        node = orm.CalculationNode().store()
        record = self.log_record
        record['dbnode_id'] = node.id
        return Log(**record), node

    def test_create_log_message(self):
        """
        Test the manual creation of a log entry
        """
        entry, node = self.create_log()

        self.assertEqual(entry.time, self.log_record['time'])
        self.assertEqual(entry.loggername, self.log_record['loggername'])
        self.assertEqual(entry.levelname, self.log_record['levelname'])
        self.assertEqual(entry.message, self.log_record['message'])
        self.assertEqual(entry.metadata, self.log_record['metadata'])
        self.assertEqual(entry.dbnode_id, self.log_record['dbnode_id'])
        self.assertEqual(entry.dbnode_id, node.id)

    def test_create_log_unserializable_metadata(self):
        """Test that unserializable data will be removed before reaching the database causing an error."""
        import functools

        def unbound_method(argument):
            return argument

        partial = functools.partial(unbound_method, 'argument')

        node = orm.CalculationNode().store()

        # An unbound method in the `args` of the metadata
        node.logger.error('problem occurred in method %s', unbound_method)

        # A partial in the `args` of the metadata
        node.logger.error('problem occurred in partial %s', partial)

        # An exception which will include an `exc_info` object
        try:
            raise ValueError
        except ValueError:
            node.logger.exception('caught an exception')

        self.assertEqual(len(Log.objects.all()), 3)

    def test_log_delete_single(self):
        """Test that a single log entry can be deleted through the collection."""
        entry, _ = self.create_log()
        log_id = entry.id

        self.assertEqual(len(Log.objects.all()), 1)

        # Deleting the entry
        Log.objects.delete(log_id)
        self.assertEqual(len(Log.objects.all()), 0)

        # Deleting a non-existing entry should raise
        with self.assertRaises(exceptions.NotExistent):
            Log.objects.delete(log_id)

    def test_log_collection_delete_all(self):
        """Test deleting all Log entries through collection"""
        count = 10
        for _ in range(count):
            self.create_log()
        log_id = Log.objects.find(limit=1)[0].id

        self.assertEqual(len(Log.objects.all()), count)

        # Delete all
        Log.objects.delete_all()

        # Checks
        self.assertEqual(len(Log.objects.all()), 0)

        with self.assertRaises(exceptions.NotExistent):
            Log.objects.delete(log_id)

        with self.assertRaises(exceptions.NotExistent):
            Log.objects.get(id=log_id)

    def test_log_collection_delete_many(self):
        """Test deleting many Logs through the collection."""
        log_ids = []
        count = 5
        for _ in range(count):
            log_, _ = self.create_log()
            log_ids.append(log_.id)
        special_log, _ = self.create_log()

        # Assert the Logs exist
        self.assertEqual(len(Log.objects.all()), count + 1)

        # Delete new Logs using filter
        filters = {'id': {'in': log_ids}}
        Log.objects.delete_many(filters=filters)

        # Make sure only the special_log Log is left
        builder = orm.QueryBuilder().append(Log, project='id')
        self.assertEqual(builder.count(), 1)
        self.assertEqual(builder.all()[0][0], special_log.id)

        for log_id in log_ids:
            with self.assertRaises(exceptions.NotExistent):
                Log.objects.delete(log_id)

            with self.assertRaises(exceptions.NotExistent):
                Log.objects.get(id=log_id)

    def test_objects_find(self):
        """Put logs in and find them"""
        node = orm.Data().store()
        for _ in range(10):
            record = self.log_record
            record['dbnode_id'] = node.id
            Log(**record)

        entries = Log.objects.all()
        self.assertEqual(10, len(entries))
        self.assertIsInstance(entries[0], Log)

    def test_find_orderby(self):
        """
        Test the order_by option of log.find
        """
        from aiida.orm.logs import OrderSpecifier, ASCENDING, DESCENDING

        node_ids = []
        for _ in range(10):
            _, node = self.create_log()
            node_ids.append(node.id)
        node_ids.sort()

        order_by = [OrderSpecifier('dbnode_id', ASCENDING)]
        res_entries = Log.objects.find(order_by=order_by)
        self.assertEqual(res_entries[0].dbnode_id, node_ids[0])

        order_by = [OrderSpecifier('dbnode_id', DESCENDING)]
        res_entries = Log.objects.find(order_by=order_by)
        self.assertEqual(res_entries[0].dbnode_id, node_ids[-1])

    def test_find_limit(self):
        """
        Test the limit option of log.find
        """
        node = orm.Data().store()
        limit = 2
        for _ in range(limit * 2):
            self.log_record['dbnode_id'] = node.id
            Log(**self.log_record)
        entries = Log.objects.find(limit=limit)
        self.assertEqual(len(entries), limit)

    def test_find_filter(self):
        """
        Test the filter option of log.find
        """
        from random import randint

        node_ids = []
        for _ in range(10):
            _, node = self.create_log()
            node_ids.append(node.id)

        node_id_of_choice = node_ids.pop(randint(0, 9))

        entries = Log.objects.find(filters={'dbnode_id': node_id_of_choice})
        self.assertEqual(len(entries), 1)
        self.assertEqual(entries[0].dbnode_id, node_id_of_choice)

    def test_db_log_handler(self):
        """
        Verify that the db log handler is attached correctly
        by firing a log message through the regular logging module
        attached to a calculation node
        """
        from aiida.orm.logs import OrderSpecifier, ASCENDING

        message = 'Testing logging of critical failure'
        node = orm.CalculationNode()

        # Firing a log for an unstored should not end up in the database
        node.logger.critical(message)

        logs = Log.objects.find()

        self.assertEqual(len(logs), 0)

        # After storing the node, logs above log level should be stored
        node.store()
        node.logger.critical(message)
        logs = Log.objects.find()

        self.assertEqual(len(logs), 1)
        self.assertEqual(logs[0].message, message)

        # Launching a second log message ensuring that both messages are correctly stored
        message2 = message + ' - Second message'
        node.logger.critical(message2)

        order_by = [OrderSpecifier('time', ASCENDING)]
        logs = Log.objects.find(order_by=order_by)

        self.assertEqual(len(logs), 2)
        self.assertEqual(logs[0].message, message)
        self.assertEqual(logs[1].message, message2)

    def test_log_querybuilder(self):
        """ Test querying for logs by joining on nodes in the QueryBuilder """
        from aiida.orm import QueryBuilder

        # Setup nodes
        log_1, calc = self.create_log()
        log_2 = Log(now(), 'loggername', logging.getLevelName(LOG_LEVEL_REPORT), calc.id, 'log message #2')
        log_3 = Log(now(), 'loggername', logging.getLevelName(LOG_LEVEL_REPORT), calc.id, 'log message #3')

        # Retrieve a node by joining on a specific log ('log_1')
        builder = QueryBuilder()
        builder.append(Log, tag='log', filters={'id': log_2.id})
        builder.append(orm.CalculationNode, with_log='log', project=['uuid'])
        nodes = builder.all()

        self.assertEqual(len(nodes), 1)
        for node in nodes:
            self.assertIn(str(node[0]), [calc.uuid])

        # Retrieve all logs for a specific node by joining on a said node
        builder = QueryBuilder()
        builder.append(orm.CalculationNode, tag='calc', filters={'id': calc.id})
        builder.append(Log, with_node='calc', project=['uuid'])
        logs = builder.all()

        self.assertEqual(len(logs), 3)
        for log in logs:
            self.assertIn(str(log[0]), [str(log_1.uuid), str(log_2.uuid), str(log_3.uuid)])

    def test_raise_wrong_metadata_type_error(self):
        """
        Test a TypeError exception is thrown with string metadata.
        Also test that metadata is correctly created.
        """
        from aiida.common import json

        # Create CalculationNode
        calc = orm.CalculationNode().store()

        # dict metadata
        correct_metadata_format = {
            'msg': 'Life is like riding a bicycle.',
            'args': '()',
            'name': 'aiida.orm.node.process.calculation.CalculationNode'
        }

        # str of dict metadata
        wrong_metadata_format = str(correct_metadata_format)

        # JSON-serialized-deserialized dict metadata
        json_metadata_format = json.loads(json.dumps(correct_metadata_format))

        # Check an error is raised when creating a Log with wrong metadata
        with self.assertRaises(TypeError):
            Log(
                now(),
                'loggername',
                logging.getLevelName(LOG_LEVEL_REPORT),
                calc.id,
                'To keep your balance, you must keep moving',
                metadata=wrong_metadata_format
            )

        # Check no error is raised when creating a Log with dict metadata
        correct_metadata_log = Log(
            now(),
            'loggername',
            logging.getLevelName(LOG_LEVEL_REPORT),
            calc.id,
            'To keep your balance, you must keep moving',
            metadata=correct_metadata_format
        )

        # Check metadata is correctly created
        self.assertEqual(correct_metadata_log.metadata, correct_metadata_format)

        # Create Log with json metadata, making sure TypeError is NOT raised
        json_metadata_log = Log(
            now(),
            'loggername',
            logging.getLevelName(LOG_LEVEL_REPORT),
            calc.id,
            'To keep your balance, you must keep moving',
            metadata=json_metadata_format
        )

        # Check metadata is correctly created
        self.assertEqual(json_metadata_log.metadata, json_metadata_format)

        # Check no error is raised if no metadata is given
        no_metadata_log = Log(
            now(),
            'loggername',
            logging.getLevelName(LOG_LEVEL_REPORT),
            calc.id,
            'To keep your balance, you must keep moving',
            metadata=None
        )

        # Check metadata is an empty dict for no_metadata_log
        self.assertEqual(no_metadata_log.metadata, {})



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Utilities that extend the basic python language."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import functools
from inspect import stack, currentframe  # pylint: disable=ungrouped-imports
import keyword
import six

# Python 3 has a nice built-in solution, but while we support python 2 we need this ugly switch
# Note that there is a difference between the PY2 and PY3 implementations as the latter allows unicode characters
# We also need to discriminate between the getargspec versus getfullargspec
if six.PY2:
    from inspect import getargspec as get_arg_spec

    def isidentifier(identifier):
        """Return whether the given string is a valid python identifier.

        :return: boolean, True if identifier is valid, False otherwise
        :raises TypeError: if identifier is not string type
        """
        import string

        type_check(identifier, six.string_types)

        if not identifier:
            return False

        if keyword.iskeyword(identifier):
            return False

        first = '_' + string.lowercase + string.uppercase  # pylint: disable=no-member
        if identifier[0] not in first:
            return False

        other = first + string.digits
        for character in identifier[1:]:
            if character not in other:
                return False

        return True

else:
    from inspect import getfullargspec as get_arg_spec  # pylint: disable=no-name-in-module

    def isidentifier(identifier):
        """Return whether the given string is a valid python identifier.

        :return: boolean, True if identifier is valid, False otherwise
        :raises TypeError: if identifier is not string type
        """
        type_check(identifier, six.string_types)
        return identifier.isidentifier() and not keyword.iskeyword(identifier)


def type_check(what, of_type, msg=None, allow_none=False):
    """Verify that object 'what' is of type 'of_type' and if not the case, raise a TypeError.

    :param what: the object to check
    :param of_type: the type (or tuple of types) to compare to
    :param msg: if specified, allows to customize the message that is passed within the TypeError exception
    :param allow_none: boolean, if True will not raise if the passed `what` is `None`
    """
    if allow_none and what is None:
        return

    if not isinstance(what, of_type):
        if msg is None:
            msg = "Got object of type '{}', expecting '{}'".format(type(what), of_type)
        raise TypeError(msg)


def protected_decorator(check=False):
    """Decorator to ensure that the decorated method is not called from outside the class hierarchy."""

    def wrap(func):  # pylint: disable=missing-docstring
        if isinstance(func, property):
            raise RuntimeError('Protected must go after @property decorator')

        args = get_arg_spec(func)[0]  # pylint: disable=deprecated-method
        if not args:
            raise RuntimeError('Can only use the protected decorator on member functions')

        # We can only perform checks if the interpreter is capable of giving
        # us the stack i.e. currentframe() produces a valid object
        if check and currentframe() is not None:

            @functools.wraps(func)
            def wrapped_fn(self, *args, **kwargs):  # pylint: disable=missing-docstring
                try:
                    calling_class = stack()[1][0].f_locals['self']
                    assert self is calling_class
                except (KeyError, AssertionError):
                    raise RuntimeError(
                        'Cannot access protected function {} from outside'
                        ' class hierarchy'.format(func.__name__)
                    )

                return func(self, *args, **kwargs)
        else:
            wrapped_fn = func

        return wrapped_fn

    return wrap


def override_decorator(check=False):
    """Decorator to signal that a method from a base class is being overridden completely."""

    def wrap(func):  # pylint: disable=missing-docstring
        if isinstance(func, property):
            raise RuntimeError('Override must go after @property decorator')

        args = get_arg_spec(func)[0]  # pylint: disable=deprecated-method
        if not args:
            raise RuntimeError('Can only use the override decorator on member functions')

        if check:

            @functools.wraps(func)
            def wrapped_fn(self, *args, **kwargs):  # pylint: disable=missing-docstring
                try:
                    getattr(super(self.__class__, self), func.__name__)
                except AttributeError:
                    raise RuntimeError('Function {} does not override a superclass method'.format(func))

                return func(self, *args, **kwargs)
        else:
            wrapped_fn = func

        return wrapped_fn

    return wrap


protected = protected_decorator(check=False)  # pylint: disable=invalid-name
override = override_decorator(check=False)  # pylint: disable=invalid-name


class classproperty(object):  # pylint: disable=too-few-public-methods,invalid-name
    """
    A class that, when used as a decorator, works as if the
    two decorators @property and @classmethod where applied together
    (i.e., the object works as a property, both for the Class and for any
    of its instance; and is called with the class cls rather than with the
    instance as its first argument).
    """

    def __init__(self, getter):
        self.getter = getter

    def __get__(self, instance, owner):
        return self.getter(owner)


class abstractclassmethod(classmethod):  # pylint: disable=too-few-public-methods, invalid-name
    """
    A decorator indicating abstract classmethods.

    Backported from python3.
    """
    __isabstractmethod__ = True

    def __init__(self, callable):  # pylint: disable=redefined-builtin
        callable.__isabstractmethod__ = True
        super(abstractclassmethod, self).__init__(callable)


class abstractstaticmethod(staticmethod):  # pylint: disable=too-few-public-methods, invalid-name
    """
    A decorator indicating abstract staticmethods.

    Similar to abstractmethod.
    Backported from python3.
    """

    __isabstractmethod__ = True

    def __init__(self, callable):  # pylint: disable=redefined-builtin
        callable.__isabstractmethod__ = True  # pylint: disable=redefined-builtin
        super(abstractstaticmethod, self).__init__(callable)


class combomethod(object):  # pylint: disable=invalid-name,too-few-public-methods
    """
    A decorator that wraps a function that can be both a classmethod or
    instancemethod and behaves accordingly::

        class A():

            @combomethod
            def do(self, **kwargs):
                isclass = kwargs.get('isclass')
                if isclass:
                    print("I am a class", self)
                else:
                    print("I am an instance", self)

        A.do()
        A().do()

        >>> I am a class __main__.A
        >>> I am an instance <__main__.A instance at 0x7f2efb116e60>

    Attention: For ease of handling, pass keyword **isclass**
    equal to True if this was called as a classmethod and False if this
    was called as an instance.
    The argument self is therefore ambiguous!
    """

    def __init__(self, method):
        self.method = method

    def __get__(self, obj=None, objtype=None):  # pylint: disable=missing-docstring

        @functools.wraps(self.method)
        def _wrapper(*args, **kwargs):  # pylint: disable=missing-docstring
            kwargs.pop('isclass', None)
            if obj is not None:
                return self.method(obj, *args, isclass=False, **kwargs)
            return self.method(objtype, *args, isclass=True, **kwargs)

        return _wrapper


class EmptyContextManager(object):  # pylint: disable=too-few-public-methods
    """
    A dummy/no-op context manager.
    """

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_value, traceback):
        pass



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
utilities for getting multi line input from the commandline
"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
import click
from aiida.common.exceptions import InputValidationError


def ensure_scripts(pre, post, summary):
    """
    A function to check if the prepend and append scripts were specified, and
    if needed ask to edit them.

    :param pre: prepend-text
    :param post: append-text
    :param summary: summary for click template
    :return:
    """
    if not pre or not post:
        return edit_pre_post(pre, post, summary)

    return pre, post


def edit_pre_post(pre=None, post=None, summary=None):
    """
    use click to call up an editor to write or edit pre / post
    execution scripts for both codes and computers
    """
    from aiida.cmdline.utils.templates import env
    template = env.get_template('prepost.bash.tpl')
    summary = summary or {}
    summary = {k: v for k, v in summary.items() if v}

    # Define a separator that will be splitting pre- and post- execution
    # parts of the submission script
    separator = '#====================================================#\n' \
                '#=               Post execution script              =#\n' \
                '#=  I am acting as a separator, do not modify me!!! =#\n' \
                '#====================================================#\n'

    content = template.render(default_pre=pre or '', separator=separator, default_post=post or '', summary=summary)
    mlinput = click.edit(content, extension='.bash')
    if mlinput:
        import re

        # Splitting the text in pre- and post- halfs
        try:
            pre, post = mlinput.split(separator)
        except ValueError as err:
            if str(err) == 'need more than 1 value to unpack':
                raise InputValidationError(
                    'Looks like you modified the '
                    'separator that should NOT be modified. Please be '
                    'careful!'
                )
            elif str(err) == 'too many values to unpack':
                raise InputValidationError(
                    'Looks like you have more than one '
                    'separator, while only one is needed '
                    '(and allowed). Please be careful!'
                )
            else:
                raise err

        # Removing all the comments starting from '#=' in both pre- and post-
        # parts
        pre = re.sub(r'(^#=.*$\n)+', '', pre, flags=re.M).strip()
        post = re.sub(r'(^#=.*$\n)+', '', post, flags=re.M).strip()
    else:
        pre, post = ('', '')
    return pre, post


def edit_comment(old_cmt=''):
    """
    call up an editor to edit comments to nodes in the database
    """
    from aiida.cmdline.utils.templates import env
    template = env.get_template('new_cmt.txt.tpl')
    content = template.render(old_comment=old_cmt)
    mlinput = click.edit(content, extension='.txt')
    if mlinput:
        import re
        regex = r'^(?!#=)(.*)$'
        cmt = '\n'.join(re.findall(regex, mlinput, flags=re.M))
        cmt = cmt.strip('\n')
    else:
        cmt = ''
    return cmt
